"""
Conditions that are optional according to the 
component configuration. These conditions, if
applicable, will be applied, but nothing will
change if they are not applicable.
"""
from yta_core.configuration.condition.mandatory import DoNeedNarration, DoNeedText, DoNeedFilenameOrUrl, DoNeedKeywords
from abc import ABC, abstractmethod


class OptionalCondition(ABC):
    """
    Class to represent a component configuration optional
    condition that will be only applied if met.
    """

    @staticmethod
    @abstractmethod
    def is_satisfied(
        configuration: 'Configuration',
        component: dict
    ):
        pass

class DoShouldBuildNarration(OptionalCondition):
    """
    Check if the configuration says that the component
    needs or can have a specific duration field and
    the provided 'component' has the fields to generate
    a voice narration so it can be done.
    """

    @staticmethod
    def is_satisfied(
        configuration: 'Configuration',
        component: dict
    ) -> bool:
        """
        Check if the configuration says that the component
        needs or can have a specific duration field and
        the provided 'component' has the fields to generate
        a voice narration so it can be done.
        """
        return (
            (
                configuration.can_have_specific_duration or
                configuration.do_need_specific_duration
            ) and
            DoNeedNarration.is_satisfied(component)
        )
    
class DoShouldBuildSpecificDuration(OptionalCondition):
    """
    Check if the configuration says that the component
    needs or can have a specific duration field and
    the provided 'component' actually need that field
    to be set.
    """

    @staticmethod
    def is_satisfied(
        configuration: 'Configuration',
        component: dict
    ) -> bool:
        """
        Check if the configuration says that the component
        needs or can have a specific duration field and
        the provided 'component' actually need that field
        to be set.
        """
        return (
            (
                configuration.can_have_specific_duration or
                configuration.do_need_specific_duration
            ) and
            DoNeedNarration.is_satisfied(component)
        )
    
class DoShouldBuildText(OptionalCondition):
    """
    Check if the configuration says that the component
    needs or can have the 'text' field set and the
    provided 'component' actually has it.
    """

    @staticmethod
    def is_satisfied(
        configuration: 'Configuration',
        component: dict
    ) -> bool:
        """
        Check if the configuration says that the component
        needs or can have the 'text' field set and the
        provided 'component' actually has it.
        """
        return (
            (
                configuration.can_have_text or
                configuration.do_need_text
            ) and
            DoNeedText.is_satisfied(component)
        )
    
class DoShouldBuildFilenameOrUrl(OptionalCondition):
    """
    Check if the configuration allows having 'filename' or
    'url' fields set and the provided 'component' actually
    has those fields set.
    """

    @staticmethod
    def is_satisfied(
        configuration: 'Configuration',
        component: dict
    ) -> bool:
        """
        Check if the configuration allows having 'filename' or
        'url' fields set and the provided 'component' actually
        has those fields set.
        """
        return (
            (
                configuration.can_have_url or
                configuration.can_have_filename or
                configuration.do_need_filename_or_url
            ) and
            DoNeedFilenameOrUrl.is_satisfied(component)
        )

class DoShouldBuildKeywords(OptionalCondition):
    """
    Check if the configuration says that the components
    needs or can have 'keywords' and it actually has that
    field set.
    """

    @staticmethod
    def is_satisfied(
        configuration: 'Configuration',
        component: dict
    ) -> bool:
        """
        Check if the configuration says that the components
        needs or can have 'keywords' and it actually has that
        field set.
        """
        return (
            (
                configuration.can_have_keywords or
                configuration.do_need_keywords
            ) and
            DoNeedKeywords.is_satisfied(component)
        )



"""
Module to include the shortcode parser instances
we need in our app.
"""
from yta_shortcodes.parser import ShortcodeParser
from yta_core.shortcodes.tag import YTAShortcodeTag
from yta_shortcodes.tag_type import ShortcodeTagType


shortcode_parser = ShortcodeParser([
    YTAShortcodeTag('meme', ShortcodeTagType.SIMPLE)
])
"""
Shortcode parser that includes all the
shortcode tags we accept in our system
and must be applied in those text that
are allowed to use the parser.
"""
empty_shortcode_parser = ShortcodeParser([])
"""
An empty shortcode parser that allows
us to discard a text if it contains
any kind of shortcode.
"""


from yta_core.builder import Builder
from yta_core.enums.field import EnhancementField, SegmentField
from yta_core.builder.dataclasses import ClippableImage
from yta_general_utils.programming.validator.parameter import ParameterValidator
from yta_general_utils.downloader import Downloader
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from moviepy import VideoFileClip
from typing import Union


__all__ = [
    'ImageBuilder'
]

class ImageBuilder(Builder):
    """
    The builder of the IMAGE type.
    """

    @staticmethod
    def build_from_enhancement(
        enhancement: dict
    ):
        """
        Build the video content from the information
        in the given 'enhancement' dict.
        """
        ParameterValidator.validate_mandatory_dict('enhancement', enhancement)

        filename = enhancement.get(EnhancementField.FILENAME.value, None)
        duration = enhancement.get(EnhancementField.DURATION.value, None)
        url = enhancement.get(EnhancementField.URL.value, None)

        return (
            ImageBuilder._build_from_filename(
                filename,
                duration
            )
            if filename is not None else
            ImageBuilder._build_from_url(
                url,
                duration
            )
        )

    @staticmethod
    def build_from_segment(
        segment: dict
    ):
        """
        Build the video content from the information
        in the given 'segment' dict.
        """
        ParameterValidator.validate_mandatory_dict('segment', segment)

        filename = segment.get(SegmentField.FILENAME.value, None)
        duration = segment.get(SegmentField.DURATION.value, None)
        url = segment.get(SegmentField.URL.value, None)

        return (
            ImageBuilder._build_from_filename(
                filename,
                duration
            )
            if filename is not None else
            ImageBuilder._build_from_url(
                url,
                duration
            )
        )
    
    @classmethod
    def _build_from_filename(
        cls,
        filename: str,
        duration: Union[float, int]
    ):
        ParameterValidator.validate_mandatory_string('filename', filename, do_accept_empty = False)
        ParameterValidator.validate_mandatory_positive_number('duration', duration)

        return VideoFileClip(
            ClippableImage(
                filename,
                duration,
                None
            ).video
        )

    @classmethod
    def _build_from_url(
        cls,
        url: str,
        duration: Union[float, int]
    ):
        ParameterValidator.validate_mandatory_string('url', url, do_accept_empty = False)
        ParameterValidator.validate_mandatory_positive_number('duration', duration)

        return VideoFileClip(
            ClippableImage(
                Downloader.download_image(
                    url,
                    Output.get_filename(None, FileTypeX.IMAGE)
                ).filename,
                duration,
                None
            ).video
        )


from yta_core.builder.youtube.enums import YoutubeChannelId
from yta_youtube.api import YoutubeAPI
from yta_youtube.enums import VideoFormatQuality, YoutubeVideoLanguage, AudioFormatQuality
from yta_youtube_enhanced.video import YoutubeVideo
from yta_general_utils.programming.decorators import singleton
from yta_general_utils.programming.validator.parameter import ParameterValidator
from yta_general_utils.random import Random
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from typing import Union

import requests


__all__ = [
    'YoutubeDownloader'
]

@singleton
class YoutubeDownloader:
    """
    Singleton class.

    This object simplifies the access to our Youtube
    provider channels. It uses the Google Youtube Data
    V3 API and, if not available, uses a community API
    to obtain the videos.

    This object is used to download memes, stock videos,
    sounds, music, etc. from our specific Youtube channels.
    """

    _ids_to_ignore: list[str]
    """
    The ids that have been previously downloaded and 
    have to be ignored in the next searches.
    """
    _do_ignore_ids: bool
    """
    A flag to indicate if previously downloaded ids 
    (that are stored in the internal '_ids_to_ignore'
    attribute) have to be ignored or not.
    """
    _service: any
    """
    Youtube service to search though the using keywords.
    """
    
    def __init__(
        self,
        do_ignore_ids: bool = True
    ):
        if not YoutubeAPI.is_youtube_token_valid():
            print('Youtube token is not valid. Please, login to get a new valid token.')
            YoutubeAPI.start_youtube_auth_flow()

        self._service = YoutubeAPI.create_youtube_service()
        self._do_ignore_ids = do_ignore_ids
        self._ids_to_ignore = []
        
    def activate_ignore_repeated(
        self
    ) -> None:
        """
        Set as True the internal flag that indicates 
        that the ids of the videos that are downloaded
        after being activated have to be ignored in 
        the next downloads, so each video is downloaded
        only once.
        """
        self._do_ignore_ids = True

    def deactivate_ignore_repeated(
        self
    ) -> None:
        """
        Set as False the internal flag that indicates 
        that the ids of the videos that are downloaded
        after being activated have to be ignored in 
        the next downloads, so each video can be 
        downloaded an unlimited amount of times.
        """
        self._do_ignore_ids = True

    def ignore_id(
        self,
        id_to_ignore: str
    ) -> None:
        """
        Add the provided 'id_to_ignore' id to the list of
        ids to ignore.
        """
        if id_to_ignore not in self._ids_to_ignore:
            self._ids_to_ignore.append(id_to_ignore)

    def reset(
        self
    ) -> None:
        """
        Empty the internal list of ids to ignore.
        """
        self._ids_to_ignore = []

    def get_video(
        self,
        id_or_url: str
    ) -> YoutubeVideo:
        """
        Returns a YoutubeVideo instance with the video
        that has the given 'id_or_url' id or url.
        """
        ParameterValidator.validate_mandatory_string('id_or_url', id_or_url, do_accept_empty = False)
        
        return YoutubeVideo(id_or_url)
    
    def download_this_video(
        self,
        youtube_video: YoutubeVideo,
        do_include_audio: bool = True,
        language: YoutubeVideoLanguage = YoutubeVideoLanguage.DEFAULT,
        output_filename: Union[str, None] = None
    ) -> Union[str, None]:
        """
        Download the provided 'youtube_video' with a maximum
        of FULL_HD (1920x1080) and store it locally. The 
        video will be downloaded with audio if the
        'do_include_audio' flag is True, and will be 
        downloaded in the given 'language' if available.

        TODO: Careful, 'language' could be not available, by
        now we are returning None in that situation.
        """
        ParameterValidator.validate_mandatory_instance_of('youtube_video', youtube_video, YoutubeVideo)
        language = YoutubeVideoLanguage.to_enum(language)
        output_filename = Output.get_filename(output_filename, FileTypeX.VIDEO)

        if (
            do_include_audio and
            not youtube_video.is_audio_language_available(language)
        ):
            return None
    
        downloaded_video = (
            youtube_video.download_with_best_quality(
                language = language,
                video_quality = VideoFormatQuality.FULL_HD,
                output_filename = output_filename
            )
            if do_include_audio else
            youtube_video.download_video_with_best_quality(
                video_quality = VideoFormatQuality.FULL_HD,
                output_filename = output_filename
            )
        )

        self.ignore_id(downloaded_video.id)

        return downloaded_video.output_filename

    def download_video(
        self,
        url: str,
        do_include_audio: bool = True,
        language: YoutubeVideoLanguage = YoutubeVideoLanguage.DEFAULT,
        output_filename: Union[str, None] = None
    ) -> Union[str, None]:
        """
        Download the video from the given 'url', including
        or not the sound depending on the 'do_include_audio'
        parameter, using the provided 'language' (that could
        be not available), and storing it locally as
        'output_filename'.

        This method will return the local filename in which
        the video has been downloaded, or None if it was not
        possible to download it.
        """
        ParameterValidator.validate_mandatory_string('url', url, do_accept_empty = False)
        ParameterValidator.validate_mandatory_bool('do_include_audio', do_include_audio)
        language = YoutubeVideoLanguage.to_enum(language)

        try:
            # TODO: This will raise an Exception if not available
            youtube_video = YoutubeVideo(url)
        except: 
            return None
        
        return self.download_this_video(
            youtube_video,
            do_include_audio = do_include_audio,
            language = language,
            output_filename = Output.get_filename(output_filename, FileTypeX.VIDEO)
        )
    
    def download_audio(
        self,
        url: str,
        language: YoutubeVideoLanguage = YoutubeVideoLanguage.DEFAULT,
        output_filename: str = None
    ):
        """
        Download the video audio from the given 'url', using
        the provided 'language' (that could be not available),
        and storing it locally as 'output_filename'.

        This method will return the local filename in which
        the audio has been downloaded, or None if it was not
        possible to download it.
        """
        ParameterValidator.validate_mandatory_string('url', url, do_accept_empty = False)
        language = YoutubeVideoLanguage.to_enum(language)
        video = YoutubeVideo(url)

        downloaded_audio = video.download_audio(
            language = language,
            output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)
        )

        self.ignore_id(downloaded_audio.id)

        return downloaded_audio.output_filename

    # TODO: Append all the 'download_x' (from channel) here below
    def download_meme_video(
        self,
        keywords: str,
        do_include_audio: bool = True,
        do_randomize: bool = False,
        output_filename: Union[str, None] = None
    ) -> Union[str, None]:
        """
        Download a meme from the specific Meme youtube channel
        including the audio if the 'do_include_audio' parameter
        is True. It will choose a random video if the
        'do_randomize' parameter is set as True, or the first
        one found if False.

        This method returns the 'output_filename' with which 
        the video has been downloaded locally, or None if no
        one was found.
        """
        ParameterValidator.validate_mandatory_string('keywords', keywords, do_accept_empty = False)
        ParameterValidator.validate_mandatory_bool('do_include_audio', do_include_audio)
        ParameterValidator.validate_mandatory_bool('do_randomize', do_randomize)

        return self._download_video(
            keywords = keywords,
            channel_id = YoutubeChannelId.MEMES,
            do_include_audio = do_include_audio,
            do_randomize = do_randomize,
            output_filename = Output.get_filename(output_filename, FileTypeX.VIDEO)
        )
    
    def download_stock_video(
        self,
        keywords: str,
        do_include_audio: bool = True,
        do_randomize: bool = False,
        output_filename: Union[str, None] = None
    ) -> Union[str, None]:
        """
        Download a stock video from the specific Stock youtube
        channel including the audio if the 'do_include_audio'
        parameter is True. It will choose a random video if the
        'do_randomize' parameter is set as True, or the first
        one found if False.

        This method returns the 'output_filename' with which 
        the video has been downloaded locally, or None if no
        one was found.
        """
        ParameterValidator.validate_mandatory_string('keywords', keywords, do_accept_empty = False)
        ParameterValidator.validate_mandatory_bool('do_include_audio', do_include_audio)
        ParameterValidator.validate_mandatory_bool('do_randomize', do_randomize)
        
        return self._download_video(
            keywords = keywords,
            channel_id = YoutubeChannelId.STOCK,
            do_include_audio = do_include_audio,
            do_randomize = do_randomize,
            output_filename = Output.get_filename(output_filename, FileTypeX.VIDEO)
        )
    
    def download_sound_audio(
        self,
        keywords: str,
        output_filename: Union[str, None] = None
    ) -> Union[str, None]:
        """
        Download a sound audio from the specific Sounds youtube
        channel.

        This method returns the 'output_filename' with which 
        the audio has been downloaded locally, or None if no
        one was found.
        """
        ParameterValidator.validate_mandatory_string('keywords', keywords, do_accept_empty = False)

        return self._download_audio(
            keywords = keywords,
            channel_id = YoutubeChannelId.SOUNDS,
            do_randomize = False,
            output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)
        )

    def download_music_audio(
        self,
        keywords: str,
        output_filename: Union[str, None] = None
    ) -> Union[str, None]:
        """
        Download a music audio from the specific Music youtube
        channel.

        This method returns the 'output_filename' with which 
        the audio has been downloaded locally, or None if no
        one was found.
        """
        ParameterValidator.validate_mandatory_string('keywords', keywords, do_accept_empty = False)

        return self._download_audio(
            keywords = keywords,
            channel_id = YoutubeChannelId.MUSIC,
            do_randomize = False,
            output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)
        )
    
    def get_stock_video(
        self,
        keywords: str,
        do_randomize: bool = False
    ) -> Union[YoutubeVideo, None]:
        """
        Get a stock video from the Stock youtube video
        channel that matches the provided 'keywords'. It
        will return the first one or a random one if the
        'do_randomize' parameter is set as True.
        """
        ParameterValidator.validate_mandatory_string('keywords', keywords, do_accept_empty = False)
        
        return self._get_video_from_keywords(keywords, YoutubeChannelId.STOCK, do_randomize)
    
    # TODO: Append all the 'download_x' (from channel) here above

    def _get_video_from_keywords(
        self,
        keywords: str,
        channel_id: YoutubeChannelId,
        do_randomize: bool = False
    ) -> Union[YoutubeVideo, None]:
        """
        Look for a video in the channel with the given
        'channel_id' id that is obtained with the also
        provided 'keywords'. The result will be the first
        one found or a random one if 'do_randomize' is 
        True.

        This method will ignore the previously downloaded
        videos only if that internal flag is activated.
        You can activate it by using the 
        'activate_ignore_ids' method.

        For internal use only.
        """
        ParameterValidator.validate_mandatory_string('keywords', keywords, do_accept_empty = False)
        channel_id = YoutubeChannelId.to_enum(channel_id)
        ParameterValidator.validate_mandatory_bool('do_randomize', do_randomize)

        youtube_videos = self._search(keywords, 25, channel_id)

        if len(youtube_videos) == 0:
            return None
        
        # Remove the previously downloaded videos
        if (
            self._do_ignore_ids and
            len(self._ids_to_ignore) > 0
        ):
            youtube_videos = [
                youtube_video
                for youtube_video in youtube_videos
                if youtube_video['id']['videoId'] not in self._ids_to_ignore
            ]

        if len(youtube_videos) == 0:
            return None
        
        return YoutubeVideo(
            youtube_videos[0]
            if not do_randomize else
            youtube_videos[Random.int_between(0, len(youtube_videos) - 1)]
        )

    def _search(
        self,
        keywords: str,
        max_results: int = 25,
        channel_id: YoutubeChannelId = None
    ) -> list[any]:
        """
        Look for videos in the channel with the given
        'channel_id' id that are obtained with the also
        provided 'keywords'. The number of results will
        be, as maximum, the given 'max_results' parameter.

        This method returns all the videos that have been
        found with those parameters.

        For internal use only.
        """
        ParameterValidator.validate_mandatory_string('keywords', keywords, do_accept_empty = False)
        channel_id = YoutubeChannelId.to_enum(channel_id)
        max_results = 25 if max_results is None else max_results

        try:
            response_videos_list = self._service.search().list(
                part = 'snippet',
                channelId = channel_id.value,
                maxResults = max_results,
                order = 'relevance',  # This is the most interesting by far, using the youtube search engine
                type = 'video',
                q = keywords
            ).execute()
        except Exception as e:
            print(e)
            # We try the collaborative known alternative that should work
            no_key_url = f'https://yt.lemnoslife.com/noKey/search?part=snippet&channelId={channel_id.value}&maxResults={str(max_results)}&order=relevance&type=video&q={keywords}&alt=json'

            try:
                response_videos_list = requests.get(no_key_url).json()
            except Exception as e:
                print(e)
                return []

        return (
            response_videos_list['items']
            if response_videos_list['pageInfo']['totalResults'] > 0 else
            []
        )
    
    def _download_video(
        self,
        keywords: str,
        channel_id: YoutubeChannelId,
        do_include_audio: bool = False,
        do_randomize: bool = False,
        output_filename: Union[str, None] = None
    ) -> Union[None, str]:
        """
        Look for a video in the channel with the given
        'channel_id' id that is obtained with the also
        provided 'keywords'. The result will be the first
        one found or a random one if 'do_randomize' is 
        True.

        This method will ignore the previously downloaded
        videos only if that internal flag is activated.
        You can activate it by using the 
        'activate_ignore_ids' method.

        This method will return None if no videos found or
        the download was not possible, or the filename
        with which the video has been downloaded if it was
        possible.

        This method forces the download to FULL HD quality
        (1920x1080) or lower if unavailable.

        For internal use only.
        """
        ParameterValidator.validate_mandatory_string('keywords', keywords, do_accept_empty = False)
        channel_id = YoutubeChannelId.to_enum(channel_id)
        ParameterValidator.validate_mandatory_bool('do_include_audio', do_include_audio)
        ParameterValidator.validate_mandatory_bool('do_randomize', do_randomize)
        
        youtube_video = self._get_video_from_keywords(keywords, channel_id, do_randomize)

        if not youtube_video:
            return None

        download_method = (
            lambda video_quality, output_filename: youtube_video.download_with_best_quality(
                video_quality = video_quality,
                output_filename = output_filename
            )
            if do_include_audio else
            lambda video_quality, output_filename: youtube_video.download_video_with_best_quality(
                video_quality = video_quality,
                output_filename = output_filename
            )
        )

        output_filename = download_method(
            VideoFormatQuality.FULL_HD,
            Output.get_filename(output_filename, FileTypeX.VIDEO)
        )

        if self._do_ignore_ids:
            self.ignore_id(youtube_video.id)

        return output_filename
    
    def _download_audio(
        self,
        keywords: str,
        channel_id: YoutubeChannelId,
        do_randomize: bool = False,
        language: YoutubeVideoLanguage = YoutubeVideoLanguage.DEFAULT,
        output_filename: Union[str, None] = None
    ) -> Union[None, str]:
        """
        Look for the audio of a video in the channel with
        the given 'channel_id' id that is obtained with
        the also provided 'keywords'. The result will be
        the first one found or a random one if 'do_randomize'
        is True.

        This method will ignore the previously downloaded
        videos only if that internal flag is activated.
        You can activate it by using the 
        'activate_ignore_ids' method.

        This method will return None if no videos found or
        the download was not possible, or the filename
        with which the video has been downloaded if it was
        possible.

        This method will download the best audio quality
        available.

        For internal use only.
        """
        ParameterValidator.validate_mandatory_string('keywords', keywords, do_accept_empty = False)
        channel_id = YoutubeChannelId.to_enum(channel_id)
        ParameterValidator.validate_mandatory_bool('do_randomize', do_randomize)
        
        youtube_video = self._get_video_from_keywords(keywords, channel_id, do_randomize)

        if not youtube_video:
            return None
        
        youtube_video.download_audio_with_best_quality(
            language = language,
            quality = AudioFormatQuality.BEST,
            output_filename = Output.get_filename(output_filename, FileTypeX.AUDIO)
        )

        if self._do_ignore_ids:
            self.ignore_id(youtube_video.id)

        return output_filename


from youtube_autonomous.experimental.enhancement.enhancement_element import EnhancementElement
from youtube_autonomous.elements.validator.element_validator import ElementValidator

from yta_core.configuration import Configuration
from yta_core.enums.component import Component
from yta_core.enums.start import ShortcodeStart
from yta_core.enums.field import ShortcodeField
from yta_core.enums.type import ShortcodeType
from yta_core.enums.mode import ShortcodeMode
from yta_core.enums.string_duration import ShortcodeStringDuration
from yta_shortcodes.tag_type import ShortcodeTagType
from yta_shortcodes.shortcode import YTAShortcode as BaseShortcode
from yta_audio.voice.transcription.objects import AudioTranscription
from typing import Union


LOWER_INDEX = -1
UPPER_INDEX = 99999

class Shortcode(BaseShortcode):
    """
    Class that represent a shortcode detected in a text, containing 
    its attributes and values and also the content if block-scoped.
    """
    
    tag: ShortcodeType
    """
    The shortcode tag that represents and identifies
    it, determining the way it will be built and
    applied.
    """
    type: ShortcodeTagType
    """
    The type of the shortcode, that identifies it
    as a simple or a block shortcode.
    """
    _start: float
    _duration: float
    _keywords: str
    _filename: str
    _url: str
    _mode: ShortcodeMode
    context: any # TODO: Do I really need this (?)
    content: Union[str, None]
    """
    The text that is between the shortcode open and
    end tag and can include shortcodes. This
    parameter makes sense if the shortcode is a
    block-scoped shortcode and will be None if it is
    a simple-scoped one.
    """
    start_previous_word_index: Union[int, None]
    """
    The index of the word that is inmediately before
    the start tag of this shortcode. Could be None if
    the shortcode is just at the begining. This index
    is considered within the text empty of shortcodes.
    """
    end_previous_word_index: Union[int, None]
    """
    The index of the word that is inmediately before
    the end tag of this shortcode. It is None if the
    shortcode is a simple one. This index is 
    considered within the text empty of shortcodes.
    """

    def __init__(
        self,
        tag: ShortcodeType,
        type: ShortcodeTagType,
        context,
        content: Union[str, None],
        attributes: dict,
        start_previous_word_index: Union[int, None] = None,
        end_previous_word_index: Union[int, None] = None
    ):
        """
        The shortcode has a 'type' and could include some 'attributes' that
        are the parameters inside the brackets, that could also be simple or
        include an associated value. If it is a block-scoped shortcode, it
        will have some 'content' inside of it.
        """
        tag = ShortcodeType.to_enum(tag)
        type = ShortcodeTagType.to_enum(type)

        self.tag = tag
        self.type = type
        self.context = context
        self.content = content
        self.start_previous_word_index = start_previous_word_index
        self.end_previous_word_index = end_previous_word_index

        # Now the more special attributes
        # TODO: This was '.config_as_shortcode' previously
        self.config = Configuration.get_configuration_by_type(self.tag)._config_as_shortcode
        
        self.start = attributes.get(ShortcodeField.START.value, None)
        self.duration = attributes.get(ShortcodeField.DURATION.value, None)
        self.keywords = attributes.get(ShortcodeField.KEYWORDS.value, '')
        self.filename = attributes.get(ShortcodeField.FILENAME.value, '')
        self.url = attributes.get(ShortcodeField.URL.value, '')
        self.mode = attributes.get(ShortcodeField.MODE.value, None)

        # TODO: Should I treat these shortcodes as Enhancement so I need
        # to make the same checks (more specific that these ones below) (?)

        # TODO: I think I have to check if string durations
        # are valid according to the ConfigurationAsShortcode

        if (
            not self.keywords and
            not self.filename and
            not self.url
        ):
            raise Exception('No "keywords" nor "filename" nor "url" sources available.')

    @property
    def start(
        self
    ):
        """
        The time moment of the current segment in which this element is
        expected to be applied.
        """
        return self._start
    
    @start.setter
    def start(
        self,
        start: Union[ShortcodeStart, int, float, None]
    ):
        self.start = (
            {
                ShortcodeTagType.SIMPLE: ShortcodeStart.BETWEEN_WORDS,
                ShortcodeTagType.BLOCK: ShortcodeStart.START_OF_FIRST_SHORTCODE_CONTENT_WORD
            }[self.type]
            if start is None else
            Component.SHORTCODE.get_start(start)
        )

    @property
    def duration(
        self
    ):
        """
        The duration of the shortcode, that it is calculated according to the
        field or to its content.
        """
        return self._duration
    
    @duration.setter
    def duration(
        self,
        duration: Union[ShortcodeStringDuration, int, float, None]
    ):
        self._duration = (
            {
                ShortcodeTagType.SIMPLE: ShortcodeStringDuration.FILE_DURATION,
                ShortcodeTagType.BLOCK: ShortcodeStringDuration.SHORTCODE_CONTENT
            }[self.type]
            if duration is None else
            Component.SHORTCODE.get_duration(duration)
        )

    @property
    def mode(
        self
    ):
        return self._mode
    
    @mode.setter
    def mode(
        self,
        mode: Union[ShortcodeMode, str]
    ):
        mode = ShortcodeMode.to_enum(mode)

        if mode not in self.config.modes:
            raise Exception(f'The provided "{mode.value}" is not a valid mode for this shortcode type "{self.type.value}" and tag "{self.tag.value}".')

        self._mode = mode
    

    # TODO: This code below was using an experimental class
    # so it can no longer be like that
    # def to_enhancement_element(
    #     self,
    #     transcription: AudioTranscription
    # ):
    #     """
    #     Turns the current shortcode to an EnhancementElement by using
    #     the provided 'transcription' and using its words to set the
    #     actual 'start' and 'duration' fields according the narration.

    #     The provided 'transcription' could be not needed if the segment
    #     is not narrated and 'start' and 'duration' fields are manually
    #     set by the user in the shortcode.
    #     """
    #     if (
    #         self.type == ShortcodeTagType.SIMPLE and
    #         self.start_previous_word_index is None
    #     ):
    #         raise Exception(f'Found {ShortcodeTagType.SIMPLE.value} shortcode without "start_previous_word_index".')
        
    #     if (
    #         self.type == ShortcodeTagType.BLOCK and
    #         (self.start_previous_word_index is None or self.end_previous_word_index is None)
    #     ):
    #         raise Exception(f'Found {ShortcodeTagType.BLOCK.value} shortcode without "start_previous_word_index" or "end_previous_word_index".')
        
    #     self.calculate_start_and_duration(transcription)

    #     # TODO: Build it
    #     enhancement_element = EnhancementElement.get_class_from_type(self.tag)(self.tag, self.start, self.duration, self.keywords, self.url, self.filename, self.mode)

    #     # TODO: Remove this below if the code above is working
    #     # if self.tag == ShortcodeType.MEME:
    #     #     enhancement_element = MemeEnhancementElement(self.tag, self.start, self.duration, self.keywords, self.url, self.filename, self.mode)
    #     # elif self.tag == ShortcodeType.SOUND:
    #     #     enhancement_element = SoundEnhancementElement(self.tag, self.start, self.duration, self.keywords, self.url, self.filename, self.mode)
    #     # elif self.tag == ShortcodeType.IMAGE:
    #     #     enhancement_element = ImageEnhancementElement(self.tag, self.start, self.duration, self.keywords, self.url, self.filename, self.mode)
    #     # elif self.tag == ShortcodeType.STICKER:
    #     #     enhancement_element = StickerEnhancementElement(self.tag, self.start, self.duration, self.keywords, self.url, self.filename, self.mode)
    #     # elif self.tag == ShortcodeType.GREEN_SCREEN:
    #     #     enhancement_element = GreenscreenEnhancementElement(self.tag, self.start, self.duration, self.keywords, self.url, self.filename, self.mode)
    #     # elif self.tag == ShortcodeType.EFFECT:
    #     #     enhancement_element = EffectEnhancementElement(self.tag, self.start, self.duration, self.keywords, self.url, self.filename, self.mode)
    #     # else:
    #     #     raise Exception(f'No valid shortcode "{self.tag}" type provided.')
    #     #     # TODO: Implement the other EnhancementElements
    #     #     enhancement_element = EnhancementElement(self.tag, self.start, self.duration, self.keywords, self.url, self.filename, self.mode)

    #     return enhancement_element


from yta_general_utils.programming.enum import YTAEnum as Enum


class _Type(Enum):
    """
    Enum class to represent the different types
    that exist in the app in a general context.
    
    For internal use only. Use a more specific
    Enum class to interact with the different
    types.
    """

    CUSTOM_STOCK = 'custom_stock'
    """
    Stock content but for our custom sources.
    """
    STOCK = 'stock'
    """
    Stock content from oficial sources.
    """
    AI_IMAGE = 'ai_image'
    """
    Content generated with AI generated images.
    """
    IMAGE = 'image'
    """
    Content generated with real images.
    """
    AI_VIDEO = 'ai_video'
    """
    Content generated with AI generated videos.
    """
    STICKER = 'sticker'
    """
    Content generated with images that are turned
    into stickers.
    """
    VIDEO = 'video'
    """
    Content generated with real videos.
    """
    SOUND = 'sound'
    """
    Content generated with real sounds.
    """
    YOUTUBE_VIDEO = 'youtube_video'
    """
    Content generated from videos obtained from
    the Youtube platform.
    """
    TEXT = 'text'
    """
    Content generated with text created with our
    own libraries.
    """
    MEME = 'meme'
    """
    Content generated with memes obtained from
    our own Youtube platform channels.
    """
    EFFECT = 'effect'
    """
    Content generated with effects made by 
    ourselves and applied to it.
    """
    PREMADE = 'premade'
    """
    Content generated by our own libraries.
    """
    GREENSCREEN = 'greenscreen'
    """
    Content generated with static and moving
    greenscreens to improve it.
    """

class SegmentType(Enum):
    """
    Enum class to represent the different types
    of a Segment, which means the content that
    will be used to fulfill that segment.
    """

    CUSTOM_STOCK = _Type.CUSTOM_STOCK.value
    """
    Stock content but for our custom sources.
    """
    STOCK = _Type.STOCK.value
    """
    Stock content from oficial sources.
    """
    AI_IMAGE = _Type.AI_IMAGE.value
    """
    Content generated with AI generated images.
    """
    IMAGE = _Type.IMAGE.value
    """
    Content generated with real images.
    """
    AI_VIDEO = _Type.AI_VIDEO.value
    """
    Content generated with AI generated videos.
    """
    STICKER = _Type.STICKER.value
    """
    Content generated with images that are turned
    into stickers.
    """
    VIDEO = _Type.VIDEO.value
    """
    Content generated with real videos.
    """
    SOUND = _Type.SOUND.value
    """
    Content generated with real sounds.
    """
    YOUTUBE_VIDEO = _Type.YOUTUBE_VIDEO.value
    """
    Content generated from videos obtained from
    the Youtube platform.
    """
    TEXT = _Type.TEXT.value
    """
    Content generated with text created with our
    own libraries.
    """
    MEME = _Type.MEME.value
    """
    Content generated with memes obtained from
    our own Youtube platform channels.
    """
    PREMADE = _Type.PREMADE.value
    """
    Content generated by our own libraries.
    """

class ShortcodeType(Enum):
    """
    Enum class to represent the different types
    of a Shortcode, which means the content that
    will be used to generate the content that a
    Shortcode indicates to be built.
    """

    CUSTOM_STOCK = _Type.CUSTOM_STOCK.value
    """
    Stock content but for our custom sources.
    """
    STOCK = _Type.STOCK.value
    """
    Stock content from oficial sources.
    """
    AI_IMAGE = _Type.AI_IMAGE.value
    """
    Content generated with AI generated images.
    """
    IMAGE = _Type.IMAGE.value
    """
    Content generated with real images.
    """
    AI_VIDEO = _Type.AI_VIDEO.value
    """
    Content generated with AI generated videos.
    """
    STICKER = _Type.STICKER.value
    """
    Content generated with images that are turned
    into stickers.
    """
    VIDEO = _Type.VIDEO.value
    """
    Content generated with real videos.
    """
    SOUND = _Type.SOUND.value
    """
    Content generated with real sounds.
    """
    YOUTUBE_VIDEO = _Type.YOUTUBE_VIDEO.value
    """
    Content generated from videos obtained from
    the Youtube platform.
    """
    TEXT = _Type.TEXT.value
    """
    Content generated with text created with our
    own libraries.
    """
    MEME = _Type.MEME.value
    """
    Content generated with memes obtained from
    our own Youtube platform channels.
    """
    PREMADE = _Type.PREMADE.value
    """
    Content generated by our own libraries.
    """
    EFFECT = _Type.EFFECT.value
    """
    Content generated with effects made by 
    ourselves and applied to it.
    """
    GREENSCREEN = _Type.GREENSCREEN.value
    """
    Content generated with static and moving
    greenscreens to improve it.
    """

class EnhancementType(Enum):
    """
    Enum class to represent the different types
    of an Enhancement, which means the content
    that will be used to generate the content
    to fulfill that Enhancement.
    """

    CUSTOM_STOCK = _Type.CUSTOM_STOCK.value
    """
    Stock content but for our custom sources.
    """
    STOCK = _Type.STOCK.value
    """
    Stock content from oficial sources.
    """
    AI_IMAGE = _Type.AI_IMAGE.value
    """
    Content generated with AI generated images.
    """
    IMAGE = _Type.IMAGE.value
    """
    Content generated with real images.
    """
    AI_VIDEO = _Type.AI_VIDEO.value
    """
    Content generated with AI generated videos.
    """
    STICKER = _Type.STICKER.value
    """
    Content generated with images that are turned
    into stickers.
    """
    VIDEO = _Type.VIDEO.value
    """
    Content generated with real videos.
    """
    SOUND = _Type.SOUND.value
    """
    Content generated with real sounds.
    """
    YOUTUBE_VIDEO = _Type.YOUTUBE_VIDEO.value
    """
    Content generated from videos obtained from
    the Youtube platform.
    """
    TEXT = _Type.TEXT.value
    """
    Content generated with text created with our
    own libraries.
    """
    MEME = _Type.MEME.value
    """
    Content generated with memes obtained from
    our own Youtube platform channels.
    """
    PREMADE = _Type.PREMADE.value
    """
    Content generated by our own libraries.
    """
    EFFECT = _Type.EFFECT.value
    """
    Content generated with effects made by 
    ourselves and applied to it.
    """
    GREENSCREEN = _Type.GREENSCREEN.value
    """
    Content generated with static and moving
    greenscreens to improve it.
    """


"""
Module to include the different configurations for
the components we are able to build within our app
so we can set the parameters we expect to have or
the ones we need to be able to build each component
according to its type.

We have 2 different types of conditions:
- Mandatory conditions, that has to be met and if
not they raise an exception.
- Optional conditions, that can be met or can be
not, but don't raise an Exception if not met.

So, we have 4 different checkings we can do:
- If we must apply a mandatory condition and it
has to because it has the fields, lets apply it.
- If we must apply a mandatory condition and it
doesn't have the fields to apply it, raise an
Exception.
- If we can apply an optional condition and it
has to, lets apply it.
- If we can apply an optional condition and it
doesn't have the fields to, just ignore it.

Our mandatory conditions start with 'do_' and 
the optional ones with 'can_'.
"""
from yta_core.enums.type import SegmentType, EnhancementType, ShortcodeType, _Type
from yta_core.enums.mode import SegmentMode, EnhancementMode, ShortcodeMode
from yta_core.enums.string_duration import SegmentStringDuration, EnhancementStringDuration, ShortcodeStringDuration
from yta_core.configuration.condition.mandatory import DoNeedFilenameOrUrl, DoNeedKeywords, DoNeedNarration, DoNeedSpecificDuration, DoNeedText
from yta_general_utils.programming.decorators import classproperty
from yta_general_utils.programming.validator import PythonValidator
from yta_general_utils.programming.validator.parameter import ParameterValidator
from abc import ABCMeta
from typing import Union
from dataclasses import dataclass


@dataclass
class ConfigurationAsSegment:
    """
    @dataclass.
    The configuration of an element as a Segment. This
    must be used to check if the parameters provided 
    are valid for this element if acting as a Segment.
    """
    
    type: SegmentType
    modes: list[SegmentMode]
    """
    A list containing the Mode values that we accept
    in this configuration as a Segment component.
    """
    default_mode: SegmentMode
    """
    The mode that will be set by default if not one
    is provided.
    """
    string_durations: list[SegmentStringDuration]
    """
    A list containing the StringDuration values that
    we accept in this configuration as a Segment
    component.
    """
    default_string_duration: SegmentStringDuration
    """
    The mode that will be set by default if not one
    is provided.
    """

    def __init__(
        self,
        type: SegmentType
    ):
        type = SegmentType.to_enum(type)

        self.type = type,
        self.modes = [], # By now we don't accept modes
        self.default_mode = None # By now we don't accept modes
        self.string_durations = [] # By now we don't accept string durations
        self.default_string_duration = None # By now we don't accept string durations

@dataclass
class ConfigurationAsEnhancement:
    """
    @dataclass
    The configuration of a component as an Enhancement.
    This must be used to check if the parameters provided 
    are valid for this component if acting as an Enhancement.
    """

    type: EnhancementType
    modes: list[EnhancementMode]
    """
    A list containing the mode values that we accept
    in this configuration as an Enhancement component.
    """
    default_mode: EnhancementMode
    """
    The mode that will be set by default if not one
    is provided.
    """
    string_durations: list[EnhancementStringDuration]
    """
    A list containing the StringDuration values that
    we accept in this configuration as an Enhancement
    component.
    """
    default_string_duration: EnhancementStringDuration
    """
    The mode that will be set by default if not one
    is provided.
    """

    def __init__(
        self,
        type: EnhancementType,
        modes: list[EnhancementMode],
        default_mode: EnhancementMode,
        string_durations: list[EnhancementStringDuration],
        default_string_duration: EnhancementStringDuration
    ):
        type = EnhancementType.to_enum(type)
        modes = [
            EnhancementMode.to_enum(mode)
            for mode in modes
        ]
        default_mode = EnhancementMode.to_enum(default_mode)
        string_durations = [
            EnhancementStringDuration.to_enum(string_duration)
            for string_duration in string_durations
        ]
        default_string_duration = EnhancementStringDuration.to_enum(default_string_duration)

        self.type = type
        self.modes = modes
        self.default_mode = default_mode
        self.string_durations = string_durations
        self.default_string_duration = default_string_duration

@dataclass
class ConfigurationAsShortcode:
    """
    @dataclass
    The configuration of an element as a Shortcode.
    This must be used to check if the parameters provided 
    are valid for this element if acting as a Shortcode.
    """

    type: ShortcodeType
    modes: list[ShortcodeMode]
    """
    A list containing the mode values that we accept
    in this configuration as a Shortode component.
    """
    default_mode: ShortcodeMode
    """
    The mode that will be set by default if no one
    is provided.
    """
    string_durations: list[ShortcodeStringDuration]
    """
    A list containing the StringDuration values that
    we accept in this configuration as a Shortcode
    component.
    """
    default_string_duration: ShortcodeStringDuration
    """
    The string duration that will be set by default
    if no one is provided.
    """

    def __init__(
        self,
        type: ShortcodeType,
        modes: list[ShortcodeMode],
        default_mode: ShortcodeMode,
        string_durations: list[ShortcodeStringDuration],
        default_string_duration: ShortcodeStringDuration
    ):
        type = ShortcodeType.to_enum(type)
        modes = [
            ShortcodeMode.to_enum(mode)
            for mode in modes
        ]
        default_mode = ShortcodeMode.to_enum(default_mode)
        string_durations = [
            ShortcodeStringDuration.to_enum(string_duration)
            for string_duration in string_durations
        ]
        default_string_duration = ShortcodeStringDuration.to_enum(default_string_duration)

        self.type = type
        self.modes = modes
        self.default_mode = default_mode
        self.string_durations = string_durations
        self.default_string_duration = default_string_duration

@dataclass
class Configuration(metaclass = ABCMeta):
    """
    Class to represent the configuration a 
    component can have according to its type.
    This must be used to check if the parameters
    provided are valid for this component.
    """
    
    _type: _Type

    _can_have_narration: bool
    """
    Can have a voice narration by setting the
    'text_to_narrate' and 'voice' fields, that
    will generate one, or by providing an
    'audio_narration_filename' that is a file
    containing a voice narration.
    """
    _do_need_voice_narration: bool
    """
    A voice narration is mandatory, so the
    'text_to_narrate' and 'voice' fields can
    be provided to generate a voice narration,
    or the 'audio_narration_filename', that is
    the file containing a voice narration.
    """
    _can_have_specific_duration: bool
    """
    The duration of this component can be set
    specifically so the element will long as
    much as the 'duration' field says.
    """
    _do_need_specific_duration: bool
    """
    The 'duration' field is mandatory in the
    situation in which the component cannot 
    have a voice narration or it can but the
    fields have not been set.
    """
    _can_have_text: bool
    """
    The 'text' field can be used to build the
    element in a specific way (determined by
    its type).
    """
    _do_need_text: bool
    """
    The 'text' field is mandatory so it must
    be set.
    """
    _can_have_filename: bool
    """
    The 'filename' field can be set to be used
    to load a file from that local stored
    filename.
    """
    _can_have_url: bool
    """
    The 'url' field can be set to be used to
    download a file from that url and use it.
    """
    _do_need_filename_or_url: bool
    """
    The 'url' or the 'filename' fields (one of
    both) are needed to be able to obtain the
    resource that need to be used to build the
    component. If both are given, the 'url' will
    be used.
    """
    _can_have_keywords: bool
    """
    The field 'keywords' can be set to use those
    keywords to look for the resource we need to
    build correctly the component (the way those
    'keywords' are used depends on the type).
    """
    _do_need_keywords: bool
    """
    The field 'keywords' is mandatory and must be
    set to be able to build the element.
    """
    _can_have_extra_parameters: bool
    """
    This field is a flag that let us know if the
    element can have some dynamic 'extra_params'
    that could be needed for the building process.
    For example, Premade and Text types could need
    extra parameters to be built ok.
    """

    _config_as_segment: Union[ConfigurationAsSegment, None]
    _config_as_enhancement: Union[ConfigurationAsEnhancement, None]
    _config_as_shortcode: Union[ConfigurationAsShortcode, None]

    @classproperty
    def type(cls):
        return cls._type

    @classproperty
    def can_have_narration(cls):
        return cls._can_have_narration
    
    @classproperty
    def do_need_voice_narration(cls):
        return cls._do_need_voice_narration
    
    @classproperty
    def can_have_specific_duration(cls):
        return cls._can_have_specific_duration
    
    @classproperty
    def do_need_specific_duration(cls):
        return cls._do_need_specific_duration
    
    @classproperty
    def can_have_text(cls):
        return cls._can_have_text
    
    @classproperty
    def do_need_text(cls):
        return cls._do_need_text
    
    @classproperty
    def can_have_filename(cls):
        return cls._can_have_filename
    
    @classproperty
    def can_have_url(cls):
        return cls._can_have_url
    
    @classproperty
    def do_need_filename_or_url(cls):
        return cls._do_need_filename_or_url

    @classproperty
    def can_have_keywords(cls):
        return cls._can_have_keywords
    
    @classproperty
    def do_need_keywords(cls):
        return cls._do_need_keywords
    
    @classproperty
    def can_have_extra_parameters(cls):
        return cls._can_have_extra_parameters
    
    @classproperty
    def config_as_segment(cls):
        return cls._config_as_segment
    
    @classproperty
    def can_be_segment(cls):
        return cls.config_as_segment is not None
    
    @classproperty
    def config_as_enhancement(cls):
        return cls._config_as_enhancement
    
    @classproperty
    def can_be_enhancement(cls):
        return cls.config_as_enhancement is not None
    
    @classproperty
    def config_as_shortcode(cls):
        return cls._config_as_shortcode
    
    @classproperty
    def can_be_shortcode(cls):
        return cls.config_as_shortcode is not None

    def __init__(
        self
    ):
        pass
    
    @staticmethod
    def get_configuration_by_type(
        type: Union[str, _Type]
        # TODO: Actually this is returning a subclass...
) -> Union['StockConfiguration', 'CustomStockConfiguration', 'AIImageConfiguration', 'ImageConfiguration', 'AIVideoConfiguration', 'VideoConfiguration', 'SoundConfiguration', 'YoutubeVideoConfiguration', 'TextConfiguration', 'MemeConfiguration', 'EffectConfiguration', 'PremadeConfiguration', 'GreenscreenConfiguration']:
        """
        Returns the configuration object that corresponds to
        the provided 'type' that need to be a valid type string
        or object.
        """
        type = (
            _Type.to_enum(type.value)
            if PythonValidator.is_enum(type) else
            _Type.to_enum(type)
        )

        return {
            _Type.STOCK: StockConfiguration,
            _Type.CUSTOM_STOCK: CustomStockConfiguration,
            _Type.AI_IMAGE: AIImageConfiguration,
            _Type.IMAGE: ImageConfiguration,
            _Type.AI_VIDEO: AIVideoConfiguration,
            _Type.VIDEO: VideoConfiguration,
            _Type.SOUND: SoundConfiguration,
            _Type.YOUTUBE_VIDEO: YoutubeVideoConfiguration,
            _Type.TEXT: TextConfiguration,
            _Type.MEME: MemeConfiguration,
            _Type.EFFECT: EffectConfiguration,
            _Type.PREMADE: PremadeConfiguration,
            _Type.GREENSCREEN: GreenscreenConfiguration
        }[type]

    def validate_component_mandatory_conditions(
        self,
        component: dict
    ):
        """
        Validate the mandatory conditions according to this
        configuration applied to the given 'component'. This
        will raise an Exception when some of those mandatory
        conditions is applicable and not met.
        """
        ParameterValidator.validate_mandatory_dict('component', component)

        if (
            self.do_need_voice_narration and
            not DoNeedNarration.is_satisfied(component)
        ):
            raise Exception('A voice narration is needed for this component and the necessary fields to build it are not set. The "audio_narration_filename" field, or the tuple "voice" and "text_to_narrate" fields are needed.')
        
        if (
            self.do_need_specific_duration and
            not DoNeedSpecificDuration.is_satisfied(component)
        ):
            raise Exception('A specific duration is needed for this component and the necessary fields to set it are not set. The "duration" field is needed, or a voice narration (built with "audio_narration_filename" or the tuple "voice" and "text_to_narrate" fields) are needed.')
        
        if (
            self.do_need_text and
            not DoNeedText.is_satisfied(component)
        ):
            raise Exception('A specific "text" field is needed and it is not set.')
        
        if (
            self.do_need_filename_or_url and
            not DoNeedFilenameOrUrl.is_satisfied(component)
        ):
            raise Exception('The "filename" field or the "url" field is needed for this component (at least one of them) and none are set.')
        
        if (
            self.do_need_keywords and
            not DoNeedKeywords.is_satisfied(component)
        ):
            raise Exception('The "keywords" field is needed for this component and it has not been set.')
        
    
@dataclass
class StockConfiguration(Configuration):
    """
    The 'Stock' element configuration, that defines if it can be used
    as a segment or as a enhancement (or both) and all the things it
    needs to work as expected.
    """
    _type = _Type.STOCK
    
    _can_have_narration = True
    _do_need_voice_narration: bool = False
    _can_have_specific_duration: bool = True
    _do_need_specific_duration: bool = False
    _can_have_text: bool = False
    _do_need_text: bool = False
    _can_have_filename: bool = False
    _can_have_url: bool = False
    _do_need_filename_or_url: bool = False
    _can_have_keywords: bool = True
    _do_need_keywords: bool = True
    _can_have_extra_parameters: bool = False

    _config_as_segment: ConfigurationAsSegment = ConfigurationAsSegment(
        type = SegmentType.STOCK,
    )
    _config_as_enhancement: ConfigurationAsEnhancement = ConfigurationAsEnhancement(
        type = EnhancementType.STOCK,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION],
        default_string_duration = EnhancementStringDuration.SEGMENT_DURATION
    )
    _config_as_shortcode: ConfigurationAsShortcode = ConfigurationAsShortcode(
        type = ShortcodeType.STOCK,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT],
        default_string_duration = ShortcodeStringDuration.SHORTCODE_CONTENT
    )

    def __init__(
        self
    ):
        pass

@dataclass
class CustomStockConfiguration(Configuration):
    """
    The 'CustomStock' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = _Type.CUSTOM_STOCK
    
    _can_have_narration = True
    _do_need_voice_narration: bool = False
    _can_have_specific_duration: bool = True
    _do_need_specific_duration: bool = False
    _can_have_text: bool = False
    _do_need_text: bool = False
    _can_have_filename: bool = False
    _can_have_url: bool = False
    _do_need_filename_or_url: bool = False
    _can_have_keywords: bool = True
    _do_need_keywords: bool = True
    _can_have_extra_parameters: bool = False

    _config_as_segment: ConfigurationAsSegment = ConfigurationAsSegment(
        type = SegmentType.CUSTOM_STOCK,
    )
    _config_as_enhancement: ConfigurationAsEnhancement = ConfigurationAsEnhancement(
        type = EnhancementType.CUSTOM_STOCK,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION],
        default_string_duration = EnhancementStringDuration.SEGMENT_DURATION
    )
    _config_as_shortcode: ConfigurationAsShortcode = ConfigurationAsShortcode(
        type = ShortcodeType.CUSTOM_STOCK,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT],
        default_string_duration = ShortcodeStringDuration.SHORTCODE_CONTENT
    )

    def __init__(
        self
    ):
        pass
    
@dataclass
class AIImageConfiguration(Configuration):
    """
    The 'AIImage' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = _Type.AI_IMAGE
    
    _can_have_narration = True
    _do_need_voice_narration: bool = False
    _can_have_specific_duration: bool = True
    _do_need_specific_duration: bool = True
    _can_have_text: bool = False
    _do_need_text: bool = False
    _can_have_filename: bool = False
    _can_have_url: bool = False
    _do_need_filename_or_url: bool = False
    _can_have_keywords: bool = True
    _do_need_keywords: bool = True
    _can_have_extra_parameters: bool = False

    _config_as_segment: ConfigurationAsSegment = ConfigurationAsSegment(
        type = SegmentType.AI_IMAGE,
    )
    _config_as_enhancement: ConfigurationAsEnhancement = ConfigurationAsEnhancement(
        type = EnhancementType.AI_IMAGE,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION],
        default_string_duration = EnhancementStringDuration.SEGMENT_DURATION
    )
    _config_as_shortcode: ConfigurationAsShortcode = ConfigurationAsShortcode(
        type = ShortcodeType.AI_IMAGE,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT],
        default_string_duration = ShortcodeStringDuration.SHORTCODE_CONTENT
    )

    def __init__(
        self
    ):
        pass

@dataclass
class ImageConfiguration(Configuration):
    """
    The 'Image' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = _Type.IMAGE
    
    _can_have_narration = True
    _do_need_voice_narration: bool = False
    _can_have_specific_duration: bool = True
    _do_need_specific_duration: bool = True
    _can_have_text: bool = False
    _do_need_text: bool = False
    _can_have_filename: bool = True
    _can_have_url: bool = True
    _do_need_filename_or_url: bool = True
    _can_have_keywords: bool = False
    _do_need_keywords: bool = False
    _can_have_extra_parameters: bool = False

    _config_as_segment: ConfigurationAsSegment = ConfigurationAsSegment(
        type = SegmentType.IMAGE,
    )
    _config_as_enhancement: ConfigurationAsEnhancement = ConfigurationAsEnhancement(
        type = EnhancementType.IMAGE,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION],
        default_string_duration = EnhancementStringDuration.SEGMENT_DURATION
    )
    _config_as_shortcode: ConfigurationAsShortcode = ConfigurationAsShortcode(
        type = ShortcodeType.IMAGE,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT],
        default_string_duration = ShortcodeStringDuration.SHORTCODE_CONTENT
    )

    def __init__(
        self
    ):
        pass

@dataclass
class AIVideoConfiguration(Configuration):
    """
    The 'AIVideo' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = _Type.AI_VIDEO
    
    _can_have_narration = True
    _do_need_voice_narration: bool = False
    _can_have_specific_duration: bool = True
    _do_need_specific_duration: bool = True
    _can_have_text: bool = False
    _do_need_text: bool = False
    _can_have_filename: bool = False
    _can_have_url: bool = False
    _do_need_filename_or_url: bool = False
    _can_have_keywords: bool = True
    _do_need_keywords: bool = True
    _can_have_extra_parameters: bool = False

    _config_as_segment: ConfigurationAsSegment = ConfigurationAsSegment(
        type = SegmentType.AI_VIDEO,
    )
    _config_as_enhancement: ConfigurationAsEnhancement = ConfigurationAsEnhancement(
        type = EnhancementType.AI_VIDEO,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION],
        default_string_duration = EnhancementStringDuration.SEGMENT_DURATION
    )
    _config_as_shortcode: ConfigurationAsShortcode = ConfigurationAsShortcode(
        type = ShortcodeType.AI_VIDEO,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT],
        default_string_duration = ShortcodeStringDuration.SHORTCODE_CONTENT
    )

    def __init__(
        self
    ):
        pass
    
@dataclass
class VideoConfiguration(Configuration):
    """
    The 'Video' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = _Type.VIDEO
    
    _can_have_narration = True
    _do_need_voice_narration: bool = False
    _can_have_specific_duration: bool = True
    _do_need_specific_duration: bool = True
    _can_have_text: bool = False
    _do_need_text: bool = False
    _can_have_filename: bool = True
    _can_have_url: bool = True
    _do_need_filename_or_url: bool = True
    _can_have_keywords: bool = False
    _do_need_keywords: bool = False
    _can_have_extra_parameters: bool = False

    _config_as_segment: ConfigurationAsSegment = ConfigurationAsSegment(
        type = SegmentType.VIDEO,
    )
    _config_as_enhancement: ConfigurationAsEnhancement = ConfigurationAsEnhancement(
        type = EnhancementType.VIDEO,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION],
        default_string_duration = EnhancementStringDuration.SEGMENT_DURATION
    )
    _config_as_shortcode: ConfigurationAsShortcode = ConfigurationAsShortcode(
        type = ShortcodeType.VIDEO,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT],
        default_string_duration = ShortcodeStringDuration.SHORTCODE_CONTENT
    )

    def __init__(
        self
    ):
        pass
    
@dataclass
class SoundConfiguration(Configuration):
    """
    The 'Sound' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = _Type.SOUND
    
    _can_have_narration = False
    _do_need_voice_narration: bool = False
    _can_have_specific_duration: bool = True
    _do_need_specific_duration: bool = False
    _can_have_text: bool = False
    _do_need_text: bool = False
    _can_have_filename: bool = True
    _can_have_url: bool = True
    _do_need_filename_or_url: bool = True
    _can_have_keywords: bool = True
    _do_need_keywords: bool = False
    _can_have_extra_parameters: bool = False

    _config_as_segment: ConfigurationAsSegment = None
    _config_as_enhancement: ConfigurationAsEnhancement = ConfigurationAsEnhancement(
        type = EnhancementType.SOUND,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION, EnhancementStringDuration.FILE_DURATION],
        default_string_duration = EnhancementStringDuration.FILE_DURATION
    )
    _config_as_shortcode: ConfigurationAsShortcode = ConfigurationAsShortcode(
        type = ShortcodeType.SOUND,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT, ShortcodeStringDuration.FILE_DURATION],
        default_string_duration = ShortcodeStringDuration.FILE_DURATION
    )

    def __init__(
        self
    ):
        pass

@dataclass
class YoutubeVideoConfiguration(Configuration):
    """
    The 'YoutubeVideo' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = _Type.YOUTUBE_VIDEO
    
    _can_have_narration = True
    _do_need_voice_narration: bool = False
    _can_have_specific_duration: bool = True
    _do_need_specific_duration: bool = True
    _can_have_text: bool = False
    _do_need_text: bool = False
    _can_have_filename: bool = False
    _can_have_url: bool = True
    _do_need_filename_or_url: bool = True
    _can_have_keywords: bool = False
    _do_need_keywords: bool = False
    _can_have_extra_parameters: bool = False

    _config_as_segment: ConfigurationAsSegment = ConfigurationAsSegment(
        type = SegmentType.YOUTUBE_VIDEO,
    )
    _config_as_enhancement: ConfigurationAsEnhancement = ConfigurationAsEnhancement(
        type = EnhancementType.YOUTUBE_VIDEO,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION],
        default_string_duration = EnhancementStringDuration.SEGMENT_DURATION
    )
    _config_as_shortcode: ConfigurationAsShortcode = ConfigurationAsShortcode(
        type = ShortcodeType.YOUTUBE_VIDEO,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT],
        default_string_duration = ShortcodeStringDuration.SHORTCODE_CONTENT
    )

    def __init__(
        self
    ):
        pass

@dataclass
class TextConfiguration(Configuration):
    """
    The 'Text' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = _Type.TEXT
    
    _can_have_narration = True
    _do_need_voice_narration: bool = False
    _can_have_specific_duration: bool = True
    _do_need_specific_duration: bool = True
    _can_have_text: bool = True
    _do_need_text: bool = True
    _can_have_filename: bool = False
    _can_have_url: bool = True
    _do_need_filename_or_url: bool = False
    _can_have_keywords: bool = True
    _do_need_keywords: bool = True
    _can_have_extra_parameters: bool = True

    _config_as_segment: ConfigurationAsSegment = ConfigurationAsSegment(
        type = SegmentType.TEXT,
    )
    _config_as_enhancement: ConfigurationAsEnhancement = ConfigurationAsEnhancement(
        type = EnhancementType.TEXT,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.OVERLAY,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION],
        default_string_duration = EnhancementStringDuration.SEGMENT_DURATION
    )
    _config_as_shortcode: ConfigurationAsShortcode = ConfigurationAsShortcode(
        type = ShortcodeType.TEXT,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.OVERLAY,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT],
        default_string_duration = ShortcodeStringDuration.SHORTCODE_CONTENT
    )

    def __init__(
        self
    ):
        pass
    
@dataclass
class MemeConfiguration(Configuration):
    """
    The 'Meme' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = _Type.MEME
    
    _can_have_narration = True
    _do_need_voice_narration: bool = False
    _can_have_specific_duration: bool = True
    _do_need_specific_duration: bool = False
    _can_have_text: bool = False
    _do_need_text: bool = False
    _can_have_filename: bool = False
    _can_have_url: bool = False
    _do_need_filename_or_url: bool = False
    _can_have_keywords: bool = True
    _do_need_keywords: bool = True
    _can_have_extra_parameters: bool = False

    _config_as_segment: ConfigurationAsSegment = ConfigurationAsSegment(
        type = SegmentType.MEME
    )
    _config_as_enhancement: ConfigurationAsEnhancement = ConfigurationAsEnhancement(
        type = EnhancementType.MEME,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.INLINE,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION, EnhancementStringDuration.FILE_DURATION],
        default_string_duration = EnhancementStringDuration.FILE_DURATION
    )
    _config_as_shortcode: ConfigurationAsShortcode = ConfigurationAsShortcode(
        type = ShortcodeType.MEME,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.INLINE,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT, ShortcodeStringDuration.FILE_DURATION],
        default_string_duration = ShortcodeStringDuration.FILE_DURATION
    )

    def __init__(
        self
    ):
        pass

@dataclass
class EffectConfiguration(Configuration):
    """
    The 'Effect' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = _Type.EFFECT
    
    _can_have_narration = False
    _do_need_voice_narration: bool = False
    _can_have_specific_duration: bool = False
    _do_need_specific_duration: bool = False
    _can_have_text: bool = False
    _do_need_text: bool = False
    _can_have_filename: bool = False
    _can_have_url: bool = False
    _do_need_filename_or_url: bool = False
    _can_have_keywords: bool = True
    _do_need_keywords: bool = True
    _can_have_extra_parameters: bool = True

    _config_as_segment: ConfigurationAsSegment = None
    _config_as_enhancement: ConfigurationAsEnhancement = ConfigurationAsEnhancement(
        type = EnhancementType.EFFECT,
        modes = [EnhancementMode.REPLACE, EnhancementMode.INLINE],
        default_mode = EnhancementMode.REPLACE,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION],
        default_string_duration = EnhancementStringDuration.SEGMENT_DURATION
    )
    _config_as_shortcode: ConfigurationAsShortcode = ConfigurationAsShortcode(
        type = ShortcodeType.EFFECT,
        modes = [ShortcodeMode.REPLACE, ShortcodeMode.INLINE],
        default_mode = ShortcodeMode.REPLACE,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT],
        default_string_duration = ShortcodeStringDuration.SHORTCODE_CONTENT
    )

    def __init__(
        self
    ):
        pass
    
@dataclass
class PremadeConfiguration(Configuration):
    """
    The 'Premade' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = _Type.PREMADE
    
    _can_have_narration = True
    _do_need_voice_narration: bool = False
    _can_have_specific_duration: bool = True
    _do_need_specific_duration: bool = True
    _can_have_text: bool = False  # dynamic parameters will be applied
    _do_need_text: bool = False
    _can_have_filename: bool = False
    _can_have_url: bool = False
    _do_need_filename_or_url: bool = False
    _can_have_keywords: bool = True
    _do_need_keywords: bool = True
    _can_have_extra_parameters: bool = True

    _config_as_segment: ConfigurationAsSegment = ConfigurationAsSegment(
        type = SegmentType.PREMADE,
    )
    _config_as_enhancement: ConfigurationAsEnhancement = ConfigurationAsEnhancement(
        type = EnhancementType.PREMADE,
        modes = [EnhancementMode.INLINE, EnhancementMode.OVERLAY],
        default_mode = EnhancementMode.INLINE,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION],
        default_string_duration = EnhancementStringDuration.SEGMENT_DURATION
    )
    _config_as_shortcode: ConfigurationAsShortcode = ConfigurationAsShortcode(
        type = ShortcodeType.PREMADE,
        modes = [ShortcodeMode.INLINE, ShortcodeMode.OVERLAY],
        default_mode = ShortcodeMode.INLINE,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT],
        default_string_duration = ShortcodeStringDuration.SHORTCODE_CONTENT
    )

    def __init__(
        self
    ):
        pass
    
@dataclass
class GreenscreenConfiguration(Configuration):
    """
    The 'Greenscreen' element configuration, that defines if it can
    be used as a segment or as a enhancement (or both) and all the
    things it needs to work as expected.
    """
    _type = _Type.GREENSCREEN
    
    _can_have_narration = False
    _do_need_voice_narration: bool = False
    _can_have_specific_duration: bool = True
    _do_need_specific_duration: bool = False
    _can_have_text: bool = False  # dynamic parameters will be applied
    _do_need_text: bool = False
    _can_have_filename: bool = True
    _can_have_url: bool = True
    _do_need_filename_or_url: bool = True
    _can_have_keywords: bool = False
    _do_need_keywords: bool = False
    _can_have_extra_parameters: bool = False

    _config_as_segment: ConfigurationAsSegment = None
    _config_as_enhancement: ConfigurationAsEnhancement = ConfigurationAsEnhancement(
        type = EnhancementType.GREENSCREEN,
        modes = [EnhancementMode.REPLACE],
        default_mode = EnhancementMode.REPLACE,
        string_durations = [EnhancementStringDuration.SEGMENT_DURATION],
        default_string_duration = EnhancementStringDuration.SEGMENT_DURATION
    )
    _config_as_shortcode: ConfigurationAsShortcode = ConfigurationAsShortcode(
        type = ShortcodeType.GREENSCREEN,
        modes = [ShortcodeMode.REPLACE],
        default_mode = ShortcodeMode.REPLACE,
        string_durations = [ShortcodeStringDuration.SHORTCODE_CONTENT],
        default_string_duration = ShortcodeStringDuration.SHORTCODE_CONTENT
    )

    def __init__(
        self
    ):
        pass



from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.validator.parameter import ParameterValidator


class _StringDuration(Enum):
    """
    Enum class to represent the different strings
    that we accept as duration values for the
    pre-processing moment. These strings will be
    transformed in a numeric value when we are able
    to do it (maybe we need to download a file, to
    create a voice narration, etc.).

    For internal use only. Use a more specific
    Enum class to interact with the different
    string durations.
    """

    SEGMENT_DURATION = 99999
    """
    The duration is expected to be the whole segment
    duration, that is only known once we have built
    it. This is, for example, useful for a music that
    we want to apply during a whole segment, or maybe
    a watermark we want to place also during the
    segment duration.
    """
    FILE_DURATION = 99998
    """
    The duration is expected to be the source file 
    (downloaded or obtained from the local system)
    duration.
    """
    SHORTCODE_CONTENT = 99997
    """
    The duration is expected to be from the begining
    of the shortcode content (start of first word) to
    the end of it (end of last word).
    """

    @staticmethod
    def to_numeric_value(
        duration: str
    ) -> int:
        """
        Transform the provided string 'duration' parameter
        to its actual numeric value according to our Enum
        declaration, or raises an Exception if not valid.
        This numeric value will be later processed by our
        system and transformed to the real duration when
        calculated.
        """
        ParameterValidator.validate_mandatory_string('duration', duration)

        duration = {
            _StringDuration.SHORTCODE_CONTENT.name.lower(): _StringDuration.SHORTCODE_CONTENT.value,
            _StringDuration.FILE_DURATION.name.lower(): _StringDuration.FILE_DURATION.value,
            _StringDuration.SEGMENT_DURATION.name.lower(): _StringDuration.SEGMENT_DURATION.value
        }.get(duration.lower(), None)

        if duration is None:
            raise Exception(f'The provided "duration" parameter {duration} is not a valid StringDuration name.')
        
        return duration
    
class SegmentStringDuration(Enum):
    """
    The string durations accepted within a Segment.
    """

    FILE_DURATION = _StringDuration.FILE_DURATION.value
    """
    The duration is expected to be the source file 
    (downloaded or obtained from the local system)
    duration.
    """

    @staticmethod
    def to_numeric_value(
        duration: str
    ):
        """
        Transform the provided string 'duration' parameter
        to its actual numeric value according to our Enum
        declaration, or raises an Exception if not valid.
        This numeric value will be later processed by our
        system and transformed to the real duration when
        calculated.
        """
        _StringDuration.to_numeric_value(duration)

class EnhancementStringDuration(Enum):
    """
    The string durations accepted within an 
    Enhancement.
    """

    FILE_DURATION = _StringDuration.FILE_DURATION.value
    """
    The duration is expected to be the source file 
    (downloaded or obtained from the local system)
    duration.
    """
    SEGMENT_DURATION = _StringDuration.SEGMENT_DURATION.value
    """
    The duration is expected to be the whole segment
    duration, that is only known once we have built
    it. This is, for example, useful for a music that
    we want to apply during a whole segment, or maybe
    a watermark we want to place also during the
    segment duration.
    """

    @staticmethod
    def to_numeric_value(
        duration: str
    ):
        """
        Transform the provided string 'duration' parameter
        to its actual numeric value according to our Enum
        declaration, or raises an Exception if not valid.
        This numeric value will be later processed by our
        system and transformed to the real duration when
        calculated.
        """
        _StringDuration.to_numeric_value(duration)

class ShortcodeStringDuration(Enum):
    """
    The string durations accepted within a Shortcode.
    """
    
    FILE_DURATION = _StringDuration.FILE_DURATION.value
    """
    The duration is expected to be the source file 
    (downloaded or obtained from the local system)
    duration.
    """
    SEGMENT_DURATION = _StringDuration.SEGMENT_DURATION.value
    """
    The duration is expected to be the whole segment
    duration, that is only known once we have built
    it. This is, for example, useful for a music that
    we want to apply during a whole segment, or maybe
    a watermark we want to place also during the
    segment duration.
    """
    SHORTCODE_CONTENT = _StringDuration.SHORTCODE_CONTENT.value
    """
    The duration is expected to be from the begining
    of the shortcode content (start of first word) to
    the end of it (end of last word).
    """

    @staticmethod
    def to_numeric_value(
        duration: str
    ):
        """
        Transform the provided string 'duration' parameter
        to its actual numeric value according to our Enum
        declaration, or raises an Exception if not valid.
        This numeric value will be later processed by our
        system and transformed to the real duration when
        calculated.
        """
        _StringDuration.to_numeric_value(duration)


from yta_general_utils.programming.enum import YTAEnum as Enum


class _Start(Enum):
    """
    The moment in which an element has to start to
    be applied.
    """

    BETWEEN_WORDS = 'between_words'
    """
    The content must be applied just in the middle 
    of two words that are dictated in the voice
    narration. This means, after the end of the 
    first word and before the start of the second
    word.
    """
    START_OF_FIRST_SHORTCODE_CONTENT_WORD = 'start_of_first_shortcode_content_word'
    """
    The content must be applied just when the first
    word of a shortcode content starts being dictated.
    """
    MIDDLE_OF_FIRST_SHORTCODE_CONTENT_WORD = 'middle_of_first_shortcode_content_word'
    """
    The content must be applied just in the middle
    of the dictation of the first word of a shortcode
    content.
    """
    END_OF_FIRST_SHORTCODE_CONTENT_WORD = 'end_of_first_shortcode_content_word'
    """
    The content must be applied just when the first
    word of a shortcode content ends being dictated.
    """

class SegmentStart(Enum):
    """
    The moment in which a Segment has to start to
    be applied.
    """

    pass

# TODO: Review this (maybe rename as it is for shortcodes, 
# not enhancement elements yet).
class EnhancementStart(Enum):
    """
    The moment in which an Enhancement has to start to
    be applied.
    """

    pass
    
class ShortcodeStart(Enum):
    """
    The moment in which a Shortcode has to start to
    be applied.
    """

    BETWEEN_WORDS = _Start.BETWEEN_WORDS.value
    """
    The content must be applied just in the middle 
    of two words that are dictated in the voice
    narration. This means, after the end of the 
    first word and before the start of the second
    word.
    """
    START_OF_FIRST_SHORTCODE_CONTENT_WORD = _Start.START_OF_FIRST_SHORTCODE_CONTENT_WORD.value
    """
    The content must be applied just when the first
    word of a shortcode content starts being dictated.
    """
    MIDDLE_OF_FIRST_SHORTCODE_CONTENT_WORD = _Start.MIDDLE_OF_FIRST_SHORTCODE_CONTENT_WORD.value
    """
    The content must be applied just in the middle
    of the dictation of the first word of a shortcode
    content.
    """
    END_OF_FIRST_SHORTCODE_CONTENT_WORD = _Start.END_OF_FIRST_SHORTCODE_CONTENT_WORD.value
    """
    The content must be applied just when the first
    word of a shortcode content ends being dictated.
    """

    @classmethod
    def get_default(cls):
        return cls.START_OF_FIRST_SHORTCODE_CONTENT_WORD


"""
TODO: Maybe I need to move this class to another
module if I'm not sure that it fits well here
"""
from yta_core.classes.enhancement_json import EnhancementJson
from dataclasses import dataclass

import json


@dataclass
class SegmentJson:
    """
    @dataclass
    The information a Segment must hold. This
    information is about a segment that is being
    processed, so its values can be different
    from the original json because they have 
    been processed.
    """

    type: str
    url: str
    text: str
    text_to_narrate: str
    voice: str
    keywords: list[str]
    filename: str
    audio_narration_filename: str
    music: str

    enhancements: list[EnhancementJson]
    extra_params: dict
    text_to_narrate_sanitized_without_shortcodes: str
    text_to_narrate_with_simplified_shortcodes: str
    text_to_narrate_sanitized: str
    duration: float
    status: str
    created_at: str

    def __init__(self):
        """
        Initializes the instance with empty string
        values for all the attributes so we can
        fill them later and individually when
        processing.
        """
        self.type = ''
        self.url = ''
        self.text = ''
        self.text_to_narrate = ''
        self.voice = ''
        self.keywords = []
        self.filename = ''
        self.audio_narration_filename = ''
        self.music = ''
        self.enhancements = []
        self.extra_params = {}
        self.text_to_narrate_sanitized_without_shortcodes = ''
        self.text_to_narrate_with_simplified_shortcodes = ''
        self.text_to_narrate_sanitized = ''
        self.duration = 0.0
        self.status = ''
        self.created_at = ''

    @property
    def as_dict(self):
        """
        Get the instance as a dictionary.
        """
        return {
            'type': self.type,
            'url': self.url,
            'text': self.text,
            'text_to_narrate': self.text_to_narrate,
            'voice': self.voice,
            'keywords': self.keywords,
            'filename': self.filename,
            'audio_narration_filename': self.audio_narration_filename,
            'music': self.music,
            'enhancements': [
                enhancement.as_dict
                for enhancement in self.enhancements
            ],
            'extra_params': self.extra_params,
            'text_to_narrate_sanitized_without_shortcodes': self.text_to_narrate_sanitized_without_shortcodes,
            'text_to_narrate_with_simplified_shortcodes': self.text_to_narrate_with_simplified_shortcodes,
            'text_to_narrate_sanitized': self.text_to_narrate_sanitized,
            'duration': self.duration,
            'status': self.status,
            'created_at': self.created_at
        }

    @property
    def as_json(self):
        """
        Get the instance as a json, transformed
        from the dictionary version.
        """
        return json.dumps(self.as_dict)


"""
A way of add some kind of structure to the information
we are handling during the building process. Each 
project has 4 different steps, so we have created
separate classes to handle it easily.

- ProjectRaw (the project is just a json within a file
that can be valid or can be not)
- ProjectValidated (the project has been validated from
a file and can be built)
- ProjectProcessed (the project has been processed and
some dynamic fields have been calculated so it is 
completely processable).
- ProjectFinished (the project has finished, but it is
like the ProjectProcessed with a 'finished' status).
"""





class Settings:
    """
    General configuration class to handle the parameters
    that modify the way the projects are built.
    """
    
    MAX_DURATION_PER_IMAGE = 5
    """
    The amount of seconds that an image can be used to build a 
    video by applying some effect on it.
    """
    MIN_DURATION_PER_IMAGE = 2
    """
    The minimum amount of seconds that an image must last when 
    treating ai_images segment building to ensure that a new
    image is needed. If the value is lower than this one, the
    previous image will stay longer to fit this short period of
    time.
    """
    MAX_DURATION_PER_YOUTUBE_SCENE = 5
    """
    The maximum number of seconds that we can extract consecutively 
    from a YouTube video (related to avoiding copyright issues).
    """
    DEFAULT_SEGMENT_PARTS_FOLDER = 'segment_files'
    """
    The folder, in the root of the project in which this library is
    executed, to maintain segment building part files in order to be
    able to work and to recover them later if something goes wrong.
    """
    DEFAULT_PROJECTS_OUTPUT_FOLDER = 'output_projects'
    """
    The folder, in the root of the project in which this library is
    executed, to store the output of the projects that are generated.
    """


"""
Module to include the different configurations for
the components we are able to build within our app
so we can set the parameters we expect to have or
the ones we need to be able to build each component
according to its type.

We have 2 different types of conditions:
- Mandatory conditions, that has to be met and if
not they raise an exception.
- Optional conditions, that can be met or can be
not, but don't raise an Exception if not met.

So, we have 4 different checkings we can do:
- If we must apply a mandatory condition and it
has to because it has the fields, lets apply it.
- If we must apply a mandatory condition and it
doesn't have the fields to apply it, raise an
Exception.
- If we can apply an optional condition and it
has to, lets apply it.
- If we can apply an optional condition and it
doesn't have the fields to, just ignore it.

Our mandatory conditions start with 'do_' and 
the optional ones with 'can_'.
"""


from yta_core.builder import Builder
from yta_core.builder.enums import EffectPremade
from yta_core.enums.field import EnhancementField
from yta_core.builder.utils import enum_name_to_class
from yta_general_utils.programming.validator.parameter import ParameterValidator


__all__ = [
    'EffectBuilder'
]

class EffectBuilder(Builder):
    """
    The builder of the EFFECT type.
    """

    @staticmethod
    def build_from_enhancement(
        enhancement: dict
    ):
        ParameterValidator.validate_mandatory_dict('enhancement', enhancement)

        keywords = enhancement.get(EnhancementField.KEYWORDS.value, None)

        return EffectBuilder.build(keywords)
    
    @staticmethod
    def build(
        keywords: str
    ):
        ParameterValidator.validate_mandatory_string('keywords', keywords, do_accept_empty = False)

        # TODO: This is not working, it does nothing but
        # obtaining the effect
        return enum_name_to_class(keywords, EffectPremade)


from bson.objectid import ObjectId
from dataclasses import dataclass


@dataclass
class MongoResult:
    """
    @dataclass
    Just a mongo database result obtained from
    a query in the database. This is a wrapper
    to know what we are receiving.
    """

    json: dict
    """
    The json information as it is stored in the
    database.
    """

    @property
    def id(
        self
    ) -> ObjectId:
        """
        Get the mongo document id from its 'json' attribute
        if existing.
        """
        return self.json.get('_id', None)

# TODO: I think we are not using this one below
@dataclass
class MongoDBProject:
    """
    @dataclass
    The representation of a project in the mongo
    database, with its '_id' ObjectId and that
    stuff.

    This is what getting a project result from 
    the database must return.
    """

    id: ObjectId
    status: str
    script: str
    segments: str

    def __init__(
        self,
        json: dict
    ):
        self.id = json['_id']
        self.status = json['status']
        self.script = json['script']
        self.segments = json['segments']

        KEYS_TO_AVOID = ['_id', 'status', 'script', 'created_at']

        # Dynamically fulfill any other field
        for key, value in json.items():
            if key not in KEYS_TO_AVOID:
                setattr(self, key, value)


"""
TODO: Maybe I need to move this class to another
module if I'm not sure that it fits well here
"""
from dataclasses import dataclass

import json


@dataclass
class EnhancementJson:
    """
    @dataclass
    The information an Enhancement must hold.
    This information is about an enhancement
    that is being processed, so its values
    can be different from the original json
    because they have been processed.
    """

    type: str
    url: str
    text: str
    text_to_narrate: str
    voice: str
    keywords: list[str]
    filename: str
    audio_narration_filename: str
    music: str

    extra_params: dict
    text_to_narrate_sanitized_without_shortcodes: str
    text_to_narrate_with_simplified_shortcodes: str
    text_to_narrate_sanitized: str
    duration: float
    status: str
    created_at: str

    def __init__(self):
        """
        Initializes the instance with empty string
        values for all the attributes so we can
        fill them later and individually when
        processing.
        """
        self.type = ''
        self.url = ''
        self.text = ''
        self.text_to_narrate = ''
        self.voice = ''
        self.keywords = []
        self.filename = ''
        self.audio_narration_filename = ''
        self.music = ''
        self.extra_params = {}
        self.text_to_narrate_sanitized_without_shortcodes = ''
        self.text_to_narrate_with_simplified_shortcodes = ''
        self.text_to_narrate_sanitized = ''
        self.duration = 0.0
        self.status = ''
        self.created_at = ''

    @property
    def as_dict(self):
        """
        Get the instance as a dictionary.
        """
        return {
            'type': self.type,
            'url': self.url,
            'text': self.text,
            'text_to_narrate': self.text_to_narrate,
            'voice': self.voice,
            'keywords': self.keywords,
            'filename': self.filename,
            'audio_narration_filename': self.audio_narration_filename,
            'music': self.music,
            'extra_params': self.extra_params,
            'text_to_narrate_sanitized_without_shortcodes': self.text_to_narrate_sanitized_without_shortcodes,
            'text_to_narrate_with_simplified_shortcodes': self.text_to_narrate_with_simplified_shortcodes,
            'text_to_narrate_sanitized': self.text_to_narrate_sanitized,
            'duration': self.duration,
            'status': self.status,
            'created_at': self.created_at
        }

    @property
    def as_json(self):
        """
        Get the instance as a json, transformed
        from the dictionary version.
        """
        return json.dumps(self.as_dict, ensure_ascii = False)


from yta_core.enums.field import SegmentField
from yta_core.enums.type import SegmentType, EnhancementType
from yta_audio.voice.generation import GoogleVoiceNarrator
from yta_general_utils.programming.parameter_obtainer import ParameterObtainer
from yta_general_utils.programming.validator.parameter import ParameterValidator
from typing import Union


class Builder:
    """
    Class to wrap the general functionality about
    building elements for a project.
    """

    # TODO: Missing 'get_subclass_by_type' method

    @staticmethod
    def build_narration(
        text: str,
        output_filename: Union[str, None]
    ) -> str:
        """
        Generates a narration file that narrates the 'text' provided and
        is stored locally as 'output_filename'. If 'text' or 
        'output_filename' fields are not provided it will raise an 
        Exception.
        """
        ParameterValidator.validate_mandatory_string('text', text, do_accept_empty = False)

        # TODO: This is not returning a FileReturn, I
        # think it is the filename str
        return GoogleVoiceNarrator.narrate(
            text = text,
            output_filename = output_filename
        )
    
    @staticmethod
    def get_parameters_from_method(
        method: callable,
        element: dict,
        parameters_to_ignore: list[str] = [],
        parameters_strictly_from_element: list[str] = [],
        parameters_strictly_from_extra_params: list[str] = []
    ) -> dict:
        """
        Extract the parameters from the given 'method' that are
        available in the also provided 'element' and 'extra_data',
        ignoring the ones in the 'parameters_to_ignore' list and
        applying the strategy that the other 2 'strictly' 
        parameters indicate.

        This method will return all the parameters that have been
        found only if all the mandatory ones were found, or will
        raise an Exception if some mandatory parameter is missing.

        The next parameters will be always ignored: 'self', 'cls',
        'args', 'kwargs'.

        The return value is dict in which the keys are the
        parameters names and the values, the values of those
        parameters.

        Params:
            parameters_to_ignore: The parameters we want to get
            ignored when obtained from the 'method' signature.

        Returns: 
            A dict in which the keys are the parameters names and
            the values are the values of those parameters.
        """
        ParameterValidator.validate_mandatory_dict('element', element)
        ParameterValidator.validate_mandatory_callable('method', method)
        ParameterValidator.validate_list_of_string('parameters_to_ignore', parameters_to_ignore)
        ParameterValidator.validate_list_of_string('parameters_strictly_from_element', parameters_strictly_from_element)
        ParameterValidator.validate_list_of_string('parameters_strictly_from_extra_params', parameters_strictly_from_extra_params)
        
        return get_parameters_for_method(
            method,
            element,
            parameters_to_ignore,
            parameters_strictly_from_element,
            parameters_strictly_from_extra_params
        )
    
def is_element_valid_for_method(
    method: callable,
    element: dict,
    parameters_to_ignore: list[str] = [],
    parameters_strictly_from_element: list[str] = [],
    parameters_strictly_from_extra_params: list[str] = []
) -> bool:
    # TODO: This method is too expensive as I'm doing
    # all the process again and again... It has to be
    # a more efficient way to do it
    try:
        get_parameters_for_method(
            method,
            element,
            parameters_to_ignore,
            parameters_strictly_from_element,
            parameters_strictly_from_extra_params
        )

        return True
    except:
        return False

def get_parameters_for_method(
    method: callable,
    element: dict,
    parameters_to_ignore: list[str] = [],
    parameters_strictly_from_element: list[str] = [],
    parameters_strictly_from_extra_params: list[str] = []
):
    """
    Extract the parameters from the given 'method' that are
    available in the also provided 'element' and 'extra_data',
    ignoring the ones in the 'parameters_to_ignore' list and
    applying the strategy that the other 2 'strictly' 
    parameters indicate.

    This method will return all the parameters that have been
    found only if all the mandatory ones were found, or will
    raise an Exception if some mandatory parameter is missing.

    The next parameters will be always ignored: 'self', 'cls',
    'args', 'kwargs'.

    The return value is dict in which the keys are the
    parameters names and the values, the values of those
    parameters.

    Params:
        parameters_to_ignore: The parameters we want to get
        ignored when obtained from the 'method' signature.

    Returns: 
        A dict in which the keys are the parameters names and
        the values are the values of those parameters.
    """
    # If arrays are None by any chance, turn into empty arrays
    parameters_to_ignore = [] if parameters_to_ignore is None else parameters_to_ignore
    parameters_strictly_from_element = [] if parameters_strictly_from_element is None else parameters_strictly_from_element
    parameters_strictly_from_extra_params = [] if parameters_strictly_from_extra_params is None else parameters_strictly_from_extra_params

    return _get_our_parameters_for_method(
        method = method,
        our_prepared_parameters = _get_parameters_from_element(
            element,
            parameters_strictly_from_element,
            parameters_strictly_from_extra_params
        ),
        parameters_to_ignore = parameters_to_ignore
    )

def _get_parameters_from_element(
    element: dict,
    parameters_strictly_from_element: list[str] = [],
    parameters_strictly_from_extra_params: list[str] = []
) -> dict:
    """
    Prepare the parameters from the given 'element'
    dict, applying the also given 'strictly'
    parameters to obtain the definitive ones that
    we will compare lately with the ones we need.

    This method returns the 'element' dict with the
    keys replaced as needed.
    """
    # If keys are duplicated in both 'strictly' arrays we
    # keep the one in 'from_element'
    parameters_strictly_from_extra_params = [
        parameter
        for parameter in parameters_strictly_from_extra_params
        if parameter not in parameters_strictly_from_element
    ]

    # TODO: Is this possible (?)
    extra_params = element.get(SegmentField.EXTRA_PARAMS.value, {})

    """
    We need to replace the 'element' attributes that are
    requested strictly from 'extra_params' and exist on it,
    and set in 'element' those 'extra_params' that are not
    in 'element' so we can use them.
    """

    for key, value in extra_params.items():
        if key in parameters_strictly_from_extra_params:
            element[key] = value
        elif (
            key not in parameters_strictly_from_element and
            key not in element
        ):
            element[key] = value

    return element

def _get_our_parameters_for_method(
    method: callable,
    our_prepared_parameters: dict,
    parameters_to_ignore: list[str] = []
):
    """
    Check if we have all the 'parameters_needed' in the
    given 'our_prepared_parameters' or not.

    The 'parameters_needed' are those obtained from the
    'get_parameters_needed_for_method', and the ones in
    'our_prepared_parameters' are the ones we obtain
    from 'prepare_element_parameters'.

    This method will raise an Exception if some of the
    'parameters_needed' is missing in
    'our_prepared_parameters'.
    """
    parameters_needed = ParameterObtainer.get_parameters_from_method(method, parameters_to_ignore)

    # 5. Raise exception if some mandatory param are missing
    missing_keys = set(parameters_needed.mandatory.as_dict.keys()) - set(our_prepared_parameters.keys())

    if len(missing_keys) > 0:
        print('We have found this parameters:')
        print(our_prepared_parameters)
        raise Exception(f'Some required parameters are missing: {", ".join(missing_keys)}')
    
    # We only want the 'parameters_needed', so remove
    # those ones from 'out_prepared_parameters' that
    # are not expected to be here
    our_prepared_parameters = {
        key: our_prepared_parameters[key]
        for key in parameters_needed.keys
        if key in our_prepared_parameters
    }
    
    # TODO: Maybe rebuild as MethodParameters to return (?)
    return our_prepared_parameters
    
# TODO: We need a method just to validate if we have all 
# of them or not. Maybe it is a good choice to create
# a dataclass to hold all this complicated operations
# with parameters
def get_builder_class(
    type: Union[SegmentType, EnhancementType]
) -> Union['AIImageBuilder', 'AIVideoBuilder', 'CustomStockBuilder', 'EffectBuilder', 'GreenscreenBuilder', 'ImageBuilder', 'MemeBuilder', 'PremadeBuilder', 'SoundBuilder', 'StockBuilder', 'TextBuilder', 'VideoBuilder', 'YoutubeVideoBuilder']:
    """
    Get the builder class associated to the
    given 'type' that must be a SegmentType
    or EnhancementType enum instance.
    """
    ParameterValidator.validate_mandatory_instance_of('type', type, [SegmentType, EnhancementType])

    from yta_core.builder.ai_image_builder import AIImageBuilder
    from yta_core.builder.ai_video_builder import AIVideoBuilder
    from yta_core.builder.custom_stock_builder import CustomStockBuilder
    from yta_core.builder.effect_builder import EffectBuilder
    from yta_core.builder.greenscreen_builder import GreenscreenBuilder
    from yta_core.builder.image_builder import ImageBuilder
    from yta_core.builder.meme_builder import MemeBuilder
    from yta_core.builder.premade_builder import PremadeBuilder
    from yta_core.builder.sound_builder import SoundBuilder
    from yta_core.builder.stock_builder import StockBuilder
    from yta_core.builder.text_builder import TextBuilder
    from yta_core.builder.video_builder import VideoBuilder
    from yta_core.builder.youtube_video_builder import YoutubeVideoBuilder

    return {
        SegmentType.AI_IMAGE: AIImageBuilder,
        EnhancementType.AI_IMAGE: AIImageBuilder,
        SegmentType.AI_VIDEO: AIVideoBuilder,
        EnhancementType.AI_VIDEO: AIVideoBuilder,
        SegmentType.CUSTOM_STOCK: CustomStockBuilder,
        EnhancementType.CUSTOM_STOCK: CustomStockBuilder,
        EnhancementType.EFFECT: EffectBuilder,
        EnhancementType.GREENSCREEN: GreenscreenBuilder,
        SegmentType.IMAGE: ImageBuilder,
        EnhancementType.IMAGE: ImageBuilder,
        SegmentType.MEME: MemeBuilder,
        EnhancementType.MEME: MemeBuilder,
        SegmentType.PREMADE: PremadeBuilder,
        EnhancementType.PREMADE: PremadeBuilder,
        SegmentType.SOUND: SoundBuilder,
        EnhancementType.SOUND: SoundBuilder,
        SegmentType.STOCK: StockBuilder,
        EnhancementType.STOCK: StockBuilder,
        SegmentType.TEXT: TextBuilder,
        EnhancementType.TEXT: TextBuilder,
        SegmentType.VIDEO: VideoBuilder,
        EnhancementType.VIDEO: VideoBuilder,
        SegmentType.YOUTUBE_VIDEO: YoutubeVideoBuilder,
        EnhancementType.YOUTUBE_VIDEO: YoutubeVideoBuilder
    }[type]
    




"""
A project that has been read from a file and
all its data is raw.
"""
from yta_core.classes.project_validated import ProjectValidated
from yta_core.classes.segment_json import SegmentJson
from yta_core.classes.enhancement_json import EnhancementJson
from yta_core.validation import SegmentJsonValidator
from yta_core.enums.status import ProjectStatus, SegmentStatus
from yta_core.enums.field import ProjectField, SegmentField, EnhancementField
from yta_core.enums.string_duration import SegmentStringDuration, EnhancementStringDuration
from yta_core.shortcodes.parser import shortcode_parser
from yta_general_utils.file.checker import FileValidator
from yta_general_utils.file.reader import FileReader
from yta_general_utils.programming.validator.parameter import ParameterValidator
from yta_general_utils.programming.validator import PythonValidator
from datetime import datetime
from dataclasses import dataclass

import copy


@dataclass
class ProjectRaw:
    """
    @dataclass
    Class to represent a raw project that has
    been read from a .json file and has to be
    validated to pass to the next building step.

    Once its been validated it will turn into 
    a ProjectValidated instance.
    """

    json: dict
    """
    The project raw json information as a dict.
    """

    @property
    def as_project_validated(
        self
    ) -> ProjectValidated:
        """
        The project but transformed into a
        ProjectValidated instance.
        """
        if not hasattr(self, '_for_database'):
            data = ProjectValidated(
                ProjectStatus.TO_START.value,
                copy.deepcopy(self.json),
                []
            )

            # TODO: This shortcode parser must be instantiated once, and this is
            # being instantiated twice in this file
            segments = []
            for segment in self.json['segments']:
                segment_data = SegmentJson()

                # We only need SegmentFields because it is being
                # read from a file, so no building process yet
                for field in SegmentField.get_all_values():
                    setattr(segment_data, field, segment[field])

                # TODO: Remove this below when working, please
                # TODO: What about 'type', 'url', 'text', etc (?)
                # type
                # url
                # text
                # keywords
                # filename
                # audio_narration_filename
                # music
                # voice
                # enhancements (?)
                # extra_params (?)

                if segment.get(SegmentField.TEXT_TO_NARRATE.value, ''):
                    # Process shortcodes in 'text_to_narrate'
                    shortcode_parser.parse(segment[SegmentField.TEXT_TO_NARRATE.value])
                    segment_data.text_to_narrate_sanitized_without_shortcodes = shortcode_parser.text_sanitized_without_shortcodes
                    segment_data.text_to_narrate_with_simplified_shortcodes = shortcode_parser.text_sanitized_with_simplified_shortcodes
                    segment_data.text_to_narrate_sanitized = shortcode_parser.text_sanitized

                # Transform string duration into its numeric value
                duration = segment.get(SegmentField.DURATION.value, None)
                if (
                    duration is not None and
                    PythonValidator.is_string(duration)
                ):
                    segment_data.duration = SegmentStringDuration.to_numeric_value(duration)
                        
                segment_data.status = SegmentStatus.TO_START.value
                segment_data.created_at = datetime.now()

                for enhancement in segment.get(SegmentField.ENHANCEMENTS.value, []):
                    enhancement_data = EnhancementJson()

                    # We only need EnhancementFields because it is
                    # being read from a file, so no building process yet
                    for field in EnhancementField.get_all_values():
                        setattr(enhancement_data, field, enhancement[field])

                    if enhancement.get(EnhancementField.TEXT_TO_NARRATE.value, ''):
                        # Process shortcodes in 'text_to_narrate'
                        shortcode_parser.parse(enhancement[EnhancementField.TEXT_TO_NARRATE.value])
                        enhancement_data.text_to_narrate_sanitized_without_shortcodes = shortcode_parser.text_sanitized_without_shortcodes
                        enhancement_data.text_to_narrate_with_simplified_shortcodes = shortcode_parser.text_sanitized_with_simplified_shortcodes
                        enhancement_data.text_to_narrate_sanitized = shortcode_parser.text_sanitized

                    # Manually handle string duration
                    enhancement_duration = enhancement.get('duration', None)
                    if (
                        enhancement_duration is not None and
                        PythonValidator.is_string(enhancement_duration)
                    ):
                        enhancement_data.duration = EnhancementStringDuration.to_numeric_value(enhancement_duration)

                    enhancement_data.status = SegmentStatus.TO_START.value
                    enhancement_data.created_at = datetime.now()

                    # Store that enhancement data
                    segment_data.enhancements.append(enhancement_data)
                segments.append(segment_data)
            data.segments = segments

            self._for_database = data

        return self._for_database
    
    def __init__(
        self,
        json: dict
    ):
        ParameterValidator.validate_mandatory_dict('json', json)

        # If project is invalid it will never be an instance
        # of it
        self.json = ProjectRaw.validate_project_json_from_file(json)

    @staticmethod
    def init_from_file(
        filename: str
    ) -> 'ProjectRaw':
        """
        Initializes a ProjectRaw instance reading the
        information from the given file 'filename'.
        """
        ParameterValidator.validate_mandatory_string('filename', filename, do_accept_empty = False)

        if not FileValidator.is_file(filename):
            raise Exception('The provided "filename" is not a valid filename.')

        return ProjectRaw(
            FileReader.read_json(filename)
        )
    
    @staticmethod
    def validate_project_json_from_file(
        json: dict
    ) -> dict:
        """
        Validate the 'json' that has been read from a file
        and raises an Exception if invalid. This method
        returns the same 'json' dict if valid.
        """
        ParameterValidator.validate_mandatory_dict('json', json)
        
        segments = json.get(ProjectField.SEGMENTS.value, None)
        if segments is None:
            raise Exception('The provided "json" dict does not contain a "segments" field.')
        
        # Lets validate the content of each segment
        for segment in segments:
            SegmentJsonValidator.validate(segment)

        return json






from yta_core.shortcodes import Shortcode
from yta_shortcodes.tag import ShortcodeTag


class YTAShortcodeTag(ShortcodeTag):
    """
    Custom Shortcode tag for Youtube Autonomous
    that uses the also custom shortcode.
    """

    def handler(
        self,
        shortcodes,
        pargs,
        kwargs,
        context,
        **extra_args
    ):
        """
        The function that handles the shortcode, fills the provided
        'shortcodes' list with a new shortcode object and all the
        required attributes and values.

        :param shortcodes: A list with the shortcodes that have been
        found previously, so we are able to store the new one.
        :param pargs: The positional arguments found in the shortcode
        tag.
        :param kwargs: The key and value arguments found in the
        shortcode tag.
        :param context: The context (I don't know what this is for...)
        :param **extra_args: Extra arguments we want to pass to this
        function.
        """
        # Fill the attributes
        attributes = {}

        for parg in pargs:
            attributes[parg] = None

        if kwargs:
            for kwarg in kwargs:
                attributes[kwarg] = kwargs[kwarg]

        # We handle the 'content' like this to avoid
        # being an strict parameter
        content = extra_args.get('content', None)

        # Here you can customize it a little bit
        # before storing the final Shortcode entity
        self.custom_handler(shortcodes, attributes, context, content)

        shortcodes.append(Shortcode(
            tag = self.name,
            type = self._type,
            context = context,
            content = content,
            attributes = attributes
        ))

        """
        We will remove all the attributes but keep 
        the shortcode tags and the content (if
        existing) to be able to detect it again in
        the text so we can extract the position for
        extra purposes
        """
        return (
            f'[{self.name}]'
            if self.is_simple_scoped else
            f'[{self.name}]{content}[/{self.name}]'
        )


from yta_general_utils.programming.enum import YTAEnum as Enum


class YoutubeChannelId(Enum):
    """
    The ids of our own Youtube channels to look
    for the specific videos we need.
    """
    
    STOCK = 'UCS49NBDOLkrahf6UgYiL4SQ'
    """
    Stock videos to be used as main part of the
    video background.
    """
    MEMES = 'UCiz62GOuKsIw3n7i3Cs4pRA'
    """
    Meme videos to enhance the experience by
    overlaying them or putting inline.
    """
    SOUNDS = 'UCF1vrXoRtVfUY05fCnmfNqw'
    """
    Sound videos to be downloaded as audio and
    used to enhance the video experience by using
    them in different specific situations.
    """
    MUSIC = 'UCDLfH57hn3K7lhmcAAqp9Mw'
    """
    Music videos to be downloaded as audio and
    used to enhance the video experience by using
    them as background music for the whole video
    or for some specific moments.
    """
    ALPHA_TRANSITIONS = 'UC8yIzZMJnFsFb40ewuUcKIA'
    """
    Alpha transitions to be downloaded and applied
    between videos with the 'alpha' transition
    effect. These transitions have no sound effect.
    """


from yta_core.builder import Builder
from yta_core.builder.dataclasses import AIImage
from yta_core.enums.field import EnhancementField, SegmentField
from yta_general_utils.programming.validator.parameter import ParameterValidator
from moviepy import concatenate_videoclips, VideoFileClip
from typing import Union


__all__ = [
    'AIImageBuilder'
]

class AIImageBuilder(Builder):
    """
    The builder of the AI_IMAGE type.
    """

    @staticmethod
    def build_from_enhancement(
        enhancement: dict
    ) -> VideoFileClip:
        """
        Build the video content from the information
        in the given 'enhancement' dict.
        """
        ParameterValidator.validate_mandatory_dict('enhancement', enhancement)

        return AIImageBuilder.build(
            enhancement.get(EnhancementField.KEYWORDS.value, None),
            enhancement.get(EnhancementField.DURATION.value, None)
        )

    @staticmethod
    def build_from_segment(
        segment: dict
    ) -> VideoFileClip:
        """
        Build the video content from the information
        in the given 'segment' dict.
        """
        ParameterValidator.validate_mandatory_dict('segment', segment)

        return AIImageBuilder.build(
            segment.get(SegmentField.KEYWORDS.value, None),
            segment.get(SegmentField.DURATION.value, None)
        )

    @staticmethod
    def build(
        keywords: str,
        duration: Union[float, int]
    ) -> VideoFileClip:
        """
        Build the video content with the given 'keywords'
        and 'duration'.
        """
        return concatenate_videoclips([
            VideoFileClip(image.video)
            for image in AIImageBuilder._get_images_array(keywords, duration)
        ])
    
    @staticmethod
    def _get_images_array(
        keywords: str,
        duration: Union[float, int],
        # TODO: Add the constant and read if from the file
        # max_duration_per_image: float = MAX_DURATION_PER_IMAGE
        max_duration_per_image: float = 5.0
    ) -> list[AIImage]:
        """
        Get a list with as many AIImage instances as needed
        to fit the provided 'duration' and according to the
        also given 'max_duration_per_image'. All those AI
        images will have been generated with the given 
        'keywords' as prompt.
        """
        ParameterValidator.validate_mandatory_string('keywords', keywords, do_accept_empty = False)
        ParameterValidator.validate_mandatory_positive_number('duration', duration, do_include_zero = False)

        images: list[AIImage] = []
        if duration > max_duration_per_image:
            number_of_images = duration // max_duration_per_image
            for _ in range(number_of_images):
                images.append(AIImage(keywords, max_duration_per_image))

            remaining_duration = duration % max_duration_per_image
            if remaining_duration:
                images[-1].duration += remaining_duration  # Add to the last image's duration
            else:
                images.append(AIImage(keywords, remaining_duration))
        else:
            images.append(AIImage(keywords, duration))

        return images



from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.text.finder import TextFinderMode as EditionManualTermMode


__all__ = [
    'EditionManualTermContext',
    'EditionManualTermField',
    'EditionManualTermMode'
]

class EditionManualTermContext(Enum):
    """
    This is the context our edition terms will have.
    The context will determine in which text topics
    we will need to apply the edition term.
    """
    
    ANY = 'any'
    """
    The term will be applied always, in any context.
    """
    
    @classmethod
    def get_default(
        cls
    ):
        """
        Get the item by default.
        """
        return cls.ANY

class EditionManualTermField(Enum):
    """
    Enum class to wrap the different fields an
    edition manual term can have.
    """

    MODE = 'mode'
    """
    The mode in which the edition manual term
    has to be searched to match the context of
    the text. Check the 'TextFinderMode' Enum
    class in TextFinder.
    """
    CONTEXT = 'context'
    """
    The context in which the the edition manual
    term should be applied, that could be a
    a topic, a category, etc.
    """
    ENHANCEMENTS = 'enhancements'
    """
    The content enhancements that must be 
    applied when this eidion manual term has
    to.
    """


from yta_multimedia.video.utils import generate_video_from_image
from yta_image.generation.ai.generator import DefaultImageGenerator
from dataclasses import dataclass
from typing import Union


@dataclass
class ClippableImage:
    """
    @dataclass
    Class to wrap the information about an Image that
    can be turned into a video clip.
    """

    duration: float
    """
    The expected duration of the AI Image when converted
    to a video clip.
    """
    _video_effect: Union[any, None]
    """
    The effect that will be applied when generating
    the video clip of the given 'duration'.

    TODO: This must be implemented in a near future
    """
    _filename: str
    """
    The filename of the image.
    """

    @property
    def image(
        self
    ) -> str:
        """
        The image filename.

        TODO: This could be a pillow Image instance.
        """
        return self._filename
    
    @property
    def video(
        self
    ) -> str:
        """
        The video generated with the also generated
        image and lasting the provided 'duration'.
        """
        if not self.image:
            return None
            
        if not hasattr(self, '_video_filename'):
            self._video_filename = generate_video_from_image(
                self.image,
                self.duration,
                fps = 60,
                # TODO: Implement the effect
                #self._video_effect
            )

        # TODO: Return it parsed as VideoFileClip (?)
        return self._video_filename
    
    def __init__(
        self,
        filename: str,
        duration: float,
        # TODO: Make this 'video_effect' more strict
        video_effect: Union[any, None] = None
    ) -> 'ClippableImage':
        # TODO: Validate it is a valid image filename (?)
        self._filename = filename
        self.duration = duration
        self._video_effect = video_effect

@dataclass
class AIImage(ClippableImage):
    """
    Class to wrap the information about an AI Image
    that is used within the AIImageBuilder.
    """

    _prompt: str
    """
    The prompt that will be used to generate the
    image.
    """
    duration: float
    """
    The expected duration of the AI Image as a clip.
    """
    _video_effect: Union[any, None]
    """
    The effect that will be applied when generating
    the video clip of the given 'duration'.
    """

    @property
    def image(
        self
    ) -> str:
        """
        The AI Image generated with the given 'prompt'.
        This method will generate the image the first
        time it is called.
        """
        if not hasattr(self, '_filename'):
            self._filename = DefaultImageGenerator().generate_image(self._prompt).filename

        # TODO: Return it parsed as Image (?)
        return self._filename
    
    def __init__(
        self,
        prompt: str,
        duration: float,
        # TODO: Make this 'video_effect' more strict
        video_effect: Union[any, None] = None
    ) -> 'AIImage':
        self._prompt = prompt
        self.duration = duration
        self._video_effect = video_effect
        self.filename = None
    
    def reset(
        self
    ):
        """
        Reset the instance to forget any previously
        generated image or video, so it will be 
        generated again the next time 'image' or
        'video' property is accessed.
        """
        self._filename = None
        self._video_filename = None



"""
Module to interact with the database in which
we store the projects and their information.
"""
from yta_core.database.dataclasses import MongoResult
from yta_core.enums.field import ProjectField, ProjectBuildingField, SegmentBuildingField
from yta_core.enums.status import ProjectStatus, SegmentStatus
from yta_mongo import MongoDatabaseHandler
from yta_mongo.clients import MongoDBClient
from yta_general_utils.programming.decorators import singleton_old
from yta_general_utils.programming.validator.parameter import ParameterValidator
from bson.objectid import ObjectId
from typing import Union


# Project read from file
# Project to store in database
# Project read from database
# Project as entity

__all__ = [
    'database_handler'
]

@singleton_old
class DatabaseHandler(MongoDatabaseHandler):
    """
    Class to interact with the database in which the
    projects are stored and its information and status
    is handled.

    The projects are stored in the database containing
    the mongo '_id'', a 'status', the 'script' (that
    is the whole json used to generate it and to
    compare other json files to check if existing), and
    the 'segments' field that is the one the app uses to
    handle the building data and this process.

    The 'script' must be preserved as it is, to be able
    to compare jsons and avoid duplicated projects, and
    the 'segments' field must be used by the app to keep
    the building process progress.
    """

    HOST = 'localhost:27017'
    DATABASE_NAME = 'youtube_autonomous'
    PROJECTS_TABLE_NAME = 'projects'

    def __init__(
        self
    ):
        super().__init__(
            self.HOST,
            self.DATABASE_NAME,
            MongoDBClient.LOCAL_MONGODB_COMPASS
        )

    @property
    def first_unfinished_project(
        self
    ) -> Union[MongoResult, None]:
        result = self.find_one_by_field(
            self.PROJECTS_TABLE_NAME,
            ProjectField.STATUS.value,
            { '$ne': ProjectStatus.FINISHED.value }
        )

        return (
            MongoResult(result)
            if result is not None else
            None
        )
    
    @property
    def unfinished_projects(
        self
    ) -> list[MongoResult]:
        results = self.find_by_field(
            self.PROJECTS_TABLE_NAME,
            ProjectField.STATUS.value,
            { '$ne': ProjectStatus.FINISHED.value }
        )

        return [
            MongoResult(result)
            for result in results
        ]
    
    def get_project_by_script(
        self,
        project: 'ProjectRaw'
    ) -> Union[MongoResult, None]:
        """
        Get the project with the provided json data
        as its 'script' field.

        The script field is where the raw project
        information is stored.
        """
        ParameterValidator.validate_mandatory_instance_of('project', project, 'ProjectRaw')

        project = self.find_one_by_field(
            self.PROJECTS_TABLE_NAME,
            'script',
            project.json
        )

        return (
            MongoResult(project)
            if project is not None else
            None
        )

    def get_project_by_id(
        self,
        id: Union[ObjectId, str]
    ) -> Union[MongoResult, None]:
        """
        Get the project with the given 'id'.
        """
        ParameterValidator.validate_mandatory_instance_of('id', id, [ObjectId, str])

        project = self.find_one_by_id(
            self.PROJECTS_TABLE_NAME,
            id
        )

        return (
            MongoResult(project)
            if project is not None else
            None
        )
    
    def insert_project(
        self,
        project: 'ProjectValidated'
    ) ->  Union[MongoResult, None]:
        """
        Insert the provided 'project' in the database.

        This method allow inserting duplicated projects.
        """
        ParameterValidator.validate_mandatory_instance_of('project', project, 'ProjectValidated')

        project = self.insert_one(
            self.PROJECTS_TABLE_NAME,
            project.as_dict
        )

        return (
            MongoResult(project)
            if project is not None else
            None
        )
    
    def update_project_field(
        self,
        project_id: Union[ObjectId, str],
        field: str,
        value: any
    ) -> Union[MongoResult, None]:
        """
        Update the given 'field' of the project with
        the also provided 'project_id' and set the 
        'value' passed as parameter.
        """
        result = self.update_one(
            table_name = self.PROJECTS_TABLE_NAME,
            id = project_id,
            field = field,
            value = value
        )

        return (
            MongoResult(result)
            if result is not None else
            None
        )
    
    # TODO: What does it return if the update is not possible (?)
    def update_project_segment_field(
        self,
        project_id: Union[ObjectId, str],
        segment_index: int,
        field: str,
        value: any
    ) -> Union[MongoResult, None]:
        """
        Update the given 'field' of the 'segment_index'
        segment, that the project with the given 
        'project_id' has, and set the 'value' passed as
        parameter.
        """
        result = self.update_one(
            table_name = self.PROJECTS_TABLE_NAME,
            id = project_id,
            field = f'segments.{str(segment_index)}.{field}',
            value = value
        )

        return (
            MongoResult(result)
            if result is not None else
            None
        )

    # TODO: What does it return if the update is not possible (?)
    def update_project_segment_enhancement_field(
        self,
        project_id: Union[ObjectId, str],
        segment_index: int,
        enhancement_index: int,
        field: str,
        value: any
    ) -> Union[MongoResult, None]:
        """
        Update the given 'field' of the enhancement in
        the 'enhancement_index' provided position, which
        is in the 'segment_index' segment of the project
        with the given  'project_id', and set the 'value' passed as
        parameter.
        """
        result = self.update_one(
            table_name = self.PROJECTS_TABLE_NAME,
            id = project_id,
            field = f'segments.{str(segment_index)}.enhancements.{str(enhancement_index)}.{field}',
            value = value
        )

        return (
            MongoResult(result)
            if result is not None else
            None
        )
    
    def update_project_segment_status(
        self,
        project_id: Union[ObjectId, str],
        segment_index: int,
        status: str
    ) -> Union[MongoResult, None]:
        """
        Update the segment with the given 'segment_index',
        that belongs to the project with the 'project_id'
        id provided and set its status field with the given
        'status' value.
        """
        return self.update_project_segment_field(
            project_id = project_id,
            segment_index = segment_index,
            field = SegmentBuildingField.STATUS.value,
            value = status
        )
    
    def set_project_segment_as_in_progress(
        self,
        project_id: Union[ObjectId, str],
        segment_index: int
    ) -> Union[MongoResult, None]:
        """
        TODO: Write doc
        """
        return self.update_project_segment_status(
            project_id = project_id,
            segment_index = segment_index,
            status = SegmentStatus.IN_PROGRESS.value
        )
    
    def set_project_segment_as_finished(
        self,
        project_id: Union[ObjectId, str],
        segment_index: int
    ) -> Union[MongoResult, None]:
        """
        TODO: Write doc
        """
        return self.update_project_segment_status(
            project_id = project_id,
            segment_index = segment_index,
            status = SegmentStatus.FINISHED.value
        )
    
    def set_project_segment_transcription(
        self,
        project_id: Union[ObjectId, str],
        segment_index: int,
        transcription: dict
    ) -> Union[MongoResult, None]:
        """
        TODO: Write doc
        """
        return self.update_project_segment_field(
            project_id = project_id,
            segment_index = segment_index,
            field = SegmentBuildingField.TRANSCRIPTION.value,
            value = transcription
        )
    
    def set_project_segment_audio_filename(
        self,
        project_id: Union[ObjectId, str],
        segment_index: int,
        audio_filename: str
    ) -> Union[MongoResult, None]:
        """
        TODO: Write doc
        """
        return self.update_project_segment_field(
            project_id = project_id,
            segment_index = segment_index,
            field = SegmentBuildingField.AUDIO_FILENAME.value,
            value = audio_filename
        )
    
    def set_project_segment_video_filename(
        self,
        project_id: Union[ObjectId, str],
        segment_index: int,
        video_filename: str
    ) -> Union[MongoResult, None]:
        """
        TODO: Write doc
        """
        return self.update_project_segment_field(
            project_id = project_id,
            segment_index = segment_index,
            field = SegmentBuildingField.VIDEO_FILENAME.value,
            value = video_filename
        )
    
    def set_project_segment_full_filename(
        self,
        project_id: Union[ObjectId, str],
        segment_index: int,
        full_filename: str
    ) -> Union[MongoResult, None]:
        """
        TODO: Write doc
        """
        return self.update_project_segment_field(
            project_id = project_id,
            segment_index = segment_index,
            field = SegmentBuildingField.FULL_FILENAME.value,
            value = full_filename
        )
    
    def set_project_music_filename(
        self,
        project_id: Union[ObjectId, str],
        music_filename: str
    ) -> Union[MongoResult, None]:
        """
        TODO: Write doc
        """
        return self.update_project_field(
            project_id = project_id,
            field = 'music_filename',
            value = music_filename
        )
    
    def set_project_as_finished(
        self,
        project_id: Union[ObjectId, str]
    ) -> Union[MongoResult, None]:
        """
        TODO: Write doc
        """
        return self.update_project_field(
            project_id = project_id,
            field = ProjectBuildingField.STATUS.value,
            value = ProjectStatus.FINISHED.value
        )
    
    def set_project_as_in_progress(
        self,
        project_id: Union[ObjectId, str]
    ) -> Union[MongoResult, None]:
        """
        TODO: Write doc
        """
        return self.update_project_field(
            project_id = project_id,
            field = ProjectBuildingField.STATUS.value,
            value = ProjectStatus.IN_PROGRESS.value
        )
    
database_handler = DatabaseHandler()


from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_multimedia.video.edition.effect.black_and_white_effect import BlackAndWhiteEffect
from yta_multimedia.video.edition.effect.blink_effect import BlinkEffect
from yta_multimedia.video.edition.effect.blur_effect import BlurEffect
from yta_multimedia.video.edition.effect.fade_in_effect import FadeInEffect
from yta_multimedia.video.edition.effect.fade_out_effect import FadeOutEffect
from yta_multimedia.video.edition.effect.flip_vertically_effect import FlipVerticallyEffect
from yta_multimedia.video.edition.effect.flip_horizontally_effect import FlipHorizontallyEffect
from yta_multimedia.video.edition.effect.multiplied_effect import MultipliedEffect
from yta_multimedia.video.edition.effect.photo_is_taken_effect import PhotoIsTakenEffect
from yta_multimedia.video.edition.effect.reversed_effect import ReversedEffect
from yta_multimedia.video.edition.effect.sad_moment_effect import SadMomentEffect
from yta_multimedia.video.edition.effect.scroll_effect import ScrollEffect
from yta_multimedia.video.edition.effect.zoom.zoom_video_effect import ZoomVideoEffect
from yta_multimedia.video.edition.effect.stop_motion_effect import StopMotionEffect
from yta_multimedia.video.edition.effect.artisanal.dynamic_postcard_effect import DynamicPostcardEffect
from yta_multimedia.video.edition.effect.artisanal.zoom_in_display_and_zoom_out_effect import ZoomInDisplayAndZoomOutEffect
from yta_multimedia.video.generation.manim.classes.text.test_text_manim_animation import TestTextManimAnimationWrapper as TestTextManimAnimation
from yta_multimedia.video.generation.manim.classes.text.magazine_text_is_written_manim_animation import MagazineTextIsWrittenManimAnimationWrapper as MagazineTextIsWrittenManimAnimation
from yta_multimedia.video.generation.manim.classes.text.magazine_text_static_manim_animation import MagazineTextStaticManimAnimationWrapper as MagazineTextStaticManimAnimation
from yta_multimedia.video.generation.manim.classes.text.rain_of_words_manim_animation import RainOfWordsManimAnimationWrapper as RainOfWordsManimAnimation
from yta_multimedia.video.generation.manim.classes.text.simple_text_manim_animation import SimpleTextManimAnimationWrapper as SimpleTextManimAnimation
from yta_multimedia.video.generation.manim.classes.text.text_word_by_word_manim_animation import TextWordByWordManimAnimationWrapper as TextWordByWordManimAnimation
from yta_multimedia.video.generation.manim.classes.text.text_triplets_manim_animation import TextTripletsManimAnimationWrapper as TextTripletsManimAnimation
from yta_multimedia.video.generation.google.google_search import GoogleSearch
from yta_multimedia.video.generation.google.youtube_search import YoutubeSearch
# from yta_multimedia.video.generation.manim.classes.video.test_video_manim_animation import TestVideoMobjectIn2DManimAnimationWrapper as TestVideoMobjectIn2DManimAnimation, TestVideoOpenGLMobjectIn2DManimAnimationWrapper as TestVideoOpenGLMobjectIn2DManimAnimation, TestVideoMobjectIn3DManimAnimationWrapper as TestVideoMobjectIn3DManimAnimation, TestVideoOpenGLMobjectIn3DManimAnimationWrapper as TestVideoOpenGLMobjectIn3DManimAnimation, TestImageOpenGLMobjectIn3DManimAnimationWrapper as TestImageOpenGLMobjectIn3DManimAnimation


class EffectPremade(Enum):
    """
    Effect premade enum class to make our effects available
    by matching the corresponding class. If the Enum exist,
    the effect is applicable.

    TODO: What about this text below: 

    These enums are pretended to be matched by their name ignoring cases,
    so feel free to use the 'get_valid_name' YTAEnum method to obtain the
    valid name to be able to intantiate it.
    """

    # TODO: This has changed completely due to moviepy 2.0.0
    # release, so I need to refactor and rethink about. Here
    # we register the accepted effects, so I think it should
    # be kept, but as I said, lets review it
    BLACK_AND_WHITE = BlackAndWhiteEffect
    BLINK = BlinkEffect
    BLUR = BlurEffect
    FADE_IN = FadeInEffect
    FADE_OUT = FadeOutEffect
    FLIP_HORIZONTALLY = FlipHorizontallyEffect
    FLIP_VERTICALLY = FlipVerticallyEffect
    MULTIPLIED = MultipliedEffect
    PHOTO_IS_TAKEN = PhotoIsTakenEffect
    REVERSED = ReversedEffect
    SAD_MOMENT = SadMomentEffect
    SCROLL = ScrollEffect
    ZOOM = ZoomVideoEffect
    STOP_MOTION = StopMotionEffect
    DYNAMIC_POSTCARD = DynamicPostcardEffect
    ZOOM_IN_OUT = ZoomInDisplayAndZoomOutEffect

class TextPremade(Enum):
    """
    Text premade enum class to make our text videos available
    by matching the corresponding class. If the Enum exist,
    the effect is applicable.

    TODO: What about this text below: 

    These enums are pretended to be matched by their name ignoring cases,
    so feel free to use the 'get_valid_name' YTAEnum method to obtain the
    valid name to be able to intantiate it.
    """
    
    TRIPLETS = TextTripletsManimAnimation
    WORD_BY_WORD = TextWordByWordManimAnimation
    SIMPLE = SimpleTextManimAnimation
    RAIN_OF_WORDS = RainOfWordsManimAnimation
    MAGAZINE_STATIC = MagazineTextStaticManimAnimation
    MAGAZINE_IS_WRITTEN = MagazineTextIsWrittenManimAnimation

    TEST = TestTextManimAnimation

class Premade(Enum):
    """
    Premade enum class to make our premade videos available
    by matching the corresponding class. If the Enum exist,
    the effect is applicable.

    TODO: What about this text below: 

    These enums are pretended to be matched by their name ignoring cases,
    so feel free to use the 'get_valid_name' YTAEnum method to obtain the
    valid name to be able to intantiate it.
    """

    GOOGLE_SEARCH = GoogleSearch
    YOUTUBE_SEARCH = YoutubeSearch
    # These below are just for testing
    # TEST_2D = TestVideoMobjectIn2DManimAnimation
    # TEST_2D_OPENGL = TestVideoOpenGLMobjectIn2DManimAnimation
    # TEST_3D = TestVideoMobjectIn3DManimAnimation
    # TEST_3D_OPENGL = TestVideoOpenGLMobjectIn3DManimAnimation
    # TEST_IMAGE_3D_OPENGL = TestImageOpenGLMobjectIn3DManimAnimation


from yta_core.enums.type import SegmentType, EnhancementType, ShortcodeType
from yta_core.enums.string_duration import SegmentStringDuration, EnhancementStringDuration, ShortcodeStringDuration
from yta_core.enums.mode import SegmentMode, EnhancementMode, ShortcodeMode
from yta_core.enums.start import SegmentStart, EnhancementStart, ShortcodeStart
from yta_core.configuration import Configuration, ConfigurationAsSegment, ConfigurationAsEnhancement, ConfigurationAsShortcode
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.validator import PythonValidator
from yta_general_utils.programming.validator.number import NumberValidator
from typing import Union


class Component(Enum):
    """
    Enum class to indicate the different componentes 
    we have in the application.
    """

    SEGMENT = 'segment'
    ENHANCEMENT = 'enhancement'
    SHORTCODE = 'shortcode'

    def get_type(
        self,
        type: Union[SegmentType, EnhancementType, ShortcodeType, str]
    ) -> Union[SegmentType, EnhancementType, ShortcodeType]:
        """
        Get the 'type' parsed as a Enum class for this
        Component, or raises an Exception if invalid.
        """
        if (
            not PythonValidator.is_string(type) and
            not PythonValidator.is_instance(type, [SegmentType, EnhancementType, ShortcodeType])
        ):
            raise Exception('The provided "type" is not an accepted value.')

        return {
            Component.SEGMENT: lambda type: SegmentType.to_enum(type),
            Component.ENHANCEMENT: lambda type: EnhancementType.to_enum(type),
            Component.SHORTCODE: lambda type: ShortcodeType.to_enum(type)
        }[self](type)

    def get_duration(
        self,
        duration: Union[SegmentStringDuration, EnhancementStringDuration, ShortcodeStringDuration, int, float, str]
    ) -> Union[SegmentStringDuration, EnhancementStringDuration, ShortcodeStringDuration, int, float, str]:
        """
        Get the 'duration' parsed as a Enum class (if a
        valid string for that) or as it is (if valid) for
        this Component, or raises an Exception if invalid.
        """
        if (
            duration is None or
            (
                not PythonValidator.is_string(duration) and
                not NumberValidator.is_positive_number(duration, do_include_zero = False) and
                not PythonValidator.is_instance(duration, SegmentStringDuration, EnhancementStringDuration, ShortcodeStringDuration)
            )
        ):
            raise Exception('The provided "duration" is not an accepted value.')
        
        return (
            {
                Component.SEGMENT: lambda duration: SegmentStringDuration.to_enum(duration),
                Component.ENHANCEMENT: lambda duration: EnhancementStringDuration.to_enum(duration),
                Component.SHORTCODE: lambda duration: ShortcodeStringDuration.to_enum(duration)
            }[self](duration)
            if not NumberValidator.is_positive_number(duration) else
            duration
        )
    
    def get_start(
        self,
        start: Union[SegmentStart, EnhancementStart, ShortcodeStart, int, float, str]
    ) -> Union[SegmentStart, EnhancementStart, ShortcodeStart]:
        """
        Get the 'start' parsed as a Enum class (if a valid
        string for that) or as it is (if valid) for this
        Component, or raises an Exception if invalid.
        """
        # TODO: I think we do not accept None as 'start'
        if (
            start is None or
            (
                not PythonValidator.is_string(start) and
                not NumberValidator.is_positive_number(start, do_include_zero = False) and
                not PythonValidator.is_instance(start, SegmentStart, EnhancementStart, ShortcodeStart)
            )
        ):
            raise Exception('The provided "duration" is not an accepted value.')
        
        return (
            {
                Component.SEGMENT: SegmentStart.to_enum(start),
                Component.ENHANCEMENT: EnhancementStart.to_enum(start),
                Component.SHORTCODE: ShortcodeStart.to_enum(start)
            }[self](start)
            if not NumberValidator.is_positive_number(start) else
            start
        )
    
    def get_mode(
        self,
        mode: Union[SegmentMode, EnhancementMode, ShortcodeMode, str],
    ) -> Union[SegmentMode, EnhancementMode, ShortcodeMode]:
        """
        Get the 'mode' parsed as a Enum class for this
        Component, or raises an Exception if invalid.
        """
        if (
            not PythonValidator.is_string(mode) and
            not PythonValidator.is_instance(mode, SegmentMode, EnhancementMode, ShortcodeMode)
        ):
            raise Exception('The provided "mode" is not an accepted value.')
        
        return {
            Component.SEGMENT: lambda mode: SegmentMode.to_enum(mode),
            Component.ENHANCEMENT: lambda mode: EnhancementMode.to_enum(mode),
            Component.SHORTCODE: lambda mode: ShortcodeMode.to_enum(mode)
        }[self](mode)

    def get_config_for_type(
        self,
        type: Union[SegmentType, EnhancementType, ShortcodeType, str]
    ) -> Union[ConfigurationAsSegment, ConfigurationAsEnhancement, ConfigurationAsShortcode]:
        """
        Get the configuration for this component for the
        specific provided 'type'.
        """
        return {
            Component.SEGMENT: lambda type: Configuration.get_configuration_by_type(type)._config_as_segment,
            Component.ENHANCEMENT: lambda type: Configuration.get_configuration_by_type(type)._config_as_enhancement,
            Component.SHORTCODE: lambda type: Configuration.get_configuration_by_type(type)._config_as_shortcode
        }[self](self.get_type(type))

    def is_mode_accepted_for_type(
        self,
        mode: Union[SegmentMode, EnhancementMode, ShortcodeMode, str, None],
        type: Union[SegmentType, EnhancementType, ShortcodeType, str]
    ) -> bool:
        """
        Check if the provided 'mode' is accepted for the
        given 'type' in this component.
        """
        type = self.get_type(type)
        mode = self.get_mode(mode)

        return mode in self.get_config_for_type(type).modes

    def is_duration_accepted_for_type(
        self,
        duration: Union[SegmentStringDuration, EnhancementStringDuration, ShortcodeStringDuration, int, float, str],
        type: Union[SegmentType, EnhancementType, ShortcodeType, str]
    ) -> bool:
        """
        Check if the provided 'duration' is accepted for
        the given 'type' in this component.
        """
        type = self.get_type(type)
        duration = self.get_duration(duration)
        config_for_type = self.get_config_for_type(type)

        string_durations = (
            config_for_type.string_durations
            if config_for_type.string_durations is not None else
            []
        )

        return (
            NumberValidator.is_positive_number(duration) or
            duration in string_durations
        )


"""
A project that can be stored in the database.
"""
from yta_core.classes.segment_json import SegmentJson
from dataclasses import dataclass

import json


@dataclass
class ProjectValidated:
    """
    @dataclass
    A project which information has been transformed
    from the raw json to the information we need to
    be stored in the database.

    It comes from a ProjectRaw instance that has been
    validated.
    """

    status: str
    """
    The current status of the project.
    """
    script: dict
    """
    The original script as it was read from the source
    file.
    """
    segments: list[SegmentJson]
    """
    The list of the segments once they've been read
    from the database.
    """

    @property
    def as_dict(
        self
    ):
        """
        Get the instance as a dictionary.
        """
        return {
            'status': self.status,
            'script': self.script,
            'segments': [
                segment.as_dict
                for segment in self.segments
            ]
        }
    
    @property
    def as_json(
        self
    ):
        """
        Get the instance as a json, transformed
        from the dictionary version.
        """
        return json.dumps(
            self.as_dict,
            ensure_ascii = False
        )
    
    #def init_from_database():


from yta_general_utils.programming.enum import YTAEnum as Enum


class _Origin(Enum):
    """
    Enum class to represent the origin of the content
    that will be displayed.
    """

    USER = 'user'
    """
    The content has been written or entered by the
    user when creating the project.
    """
    TEXT_TO_NARRATE_SHORTCODE = 'text_to_narrate_shortcode'
    """
    The content has been extracted automatically from
    the text narration that had some shortcodes manually
    written on it.
    """
    EDITION_MANUAL = 'edition_manual'
    """
    The content has been generated by applying the edition
    manual on the narration transcription.
    """
    AI_GENERATED = 'ai_generated'
    """
    The content has been generated by an AI that has
    interpreted the topic, the intention or something in
    the content that felt appropiate to include it as a
    way to improve it.
    """

class SegmentOrigin(Enum):
    """
    The origin of the content of a Segment.
    """

    USER = _Origin.USER.value
    """
    The content has been written or entered by the
    user when creating the project.
    """
    AI_GENERATED = _Origin.AI_GENERATED.value
    """
    The content has been generated by an AI that has
    interpreted the topic, the intention or something in
    the content that felt appropiate to include it as a
    way to improve it.
    """

class EnhancementOrigin(Enum):
    """
    The origin of the content of an Enhancement.
    """

    USER = _Origin.USER.value
    """
    The content has been written or entered by the
    user when creating the project.
    """
    TEXT_TO_NARRATE_SHORTCODE = _Origin.TEXT_TO_NARRATE_SHORTCODE.value
    """
    The content has been extracted automatically from
    the text narration that had some shortcodes manually
    written on it.
    """
    EDITION_MANUAL = _Origin.EDITION_MANUAL.value
    """
    The content has been generated by applying the edition
    manual on the narration transcription.
    """
    AI_GENERATED = _Origin.AI_GENERATED.value
    """
    The content has been generated by an AI that has
    interpreted the topic, the intention or something in
    the content that felt appropiate to include it as a
    way to improve it.
    """


from yta_general_utils.programming.enum import YTAEnum as Enum


class _Mode(Enum):
    """
    The mode in which the content has to be
    displayed when built.
    """

    INLINE = 'inline'
    """
    The content will be displayed in this 'inline'
    mode when it will interrupt the main video to
    be played and then the main video will continue.
    This mode modifies the clip duration so we need
    to recalculate the other objects start times.
    """
    OVERLAY = 'overlay'
    """
    The content will be displayed in this 'overlay'
    mode when it will be shown in the foreground of
    the main clip but changing not its duration. We
    don't need to recalculate any other duration.
    """
    REPLACE = 'replace'
    """
    The content will be displayed in this 'replace'
    mode when the original video is replaced by this
    content in the specific part in which it has to
    be displayed. This means that the original video
    part will be replaced by this part, but the 
    duration will remain the same as it is a 
    replacement of just a part of its content.
    """

class SegmentMode(Enum):
    """
    The modes in which a Segment can be built
    and displayed.
    """
    pass

class EnhancementMode(Enum):
    """
    The modes in which an Enhancement can be
    built and displayed.
    """

    INLINE = _Mode.INLINE.value
    """
    The content will be displayed in this 'inline'
    mode when it will interrupt the main video to
    be played and then the main video will continue.
    This mode modifies the clip duration so we need
    to recalculate the other objects start times.
    """
    OVERLAY = _Mode.OVERLAY.value
    """
    The content will be displayed in this 'overlay'
    mode when it will be shown in the foreground of
    the main clip but changing not its duration. We
    don't need to recalculate any other duration.
    """
    REPLACE = _Mode.REPLACE.value
    """
    The content will be displayed in this 'replace'
    mode when the original video is replaced by this
    content in the specific part in which it has to
    be displayed. This means that the original video
    part will be replaced by this part, but the 
    duration will remain the same as it is a 
    replacement of just a part of its content.
    """

    @classmethod
    def get_default(cls):
        """
        Returns the default enum of this list. This value will be used when
        no valid value is found.
        """
        return cls.INLINE
    
class ShortcodeMode(Enum):
    """
    The modes in which a Shortcode can be
    built and displayed.
    """

    INLINE = _Mode.INLINE.value
    """
    The content will be displayed in this 'inline'
    mode when it will interrupt the main video to
    be played and then the main video will continue.
    This mode modifies the clip duration so we need
    to recalculate the other objects start times.
    """
    OVERLAY = _Mode.OVERLAY.value
    """
    The content will be displayed in this 'overlay'
    mode when it will be shown in the foreground of
    the main clip but changing not its duration. We
    don't need to recalculate any other duration.
    """
    REPLACE = _Mode.REPLACE.value
    """
    The content will be displayed in this 'replace'
    mode when the original video is replaced by this
    content in the specific part in which it has to
    be displayed. This means that the original video
    part will be replaced by this part, but the 
    duration will remain the same as it is a 
    replacement of just a part of its content.
    """

    @classmethod
    def get_default(cls):
        """
        Returns the default enum of this list. This value will be used when
        no valid value is found.
        """
        return cls.INLINE